import re
from typing import (
    Any,
    Dict,
    Literal,
    Optional,
    Tuple,
)

import marsh


slice_pattern = re.compile(
    r'^\s*(-?\s*\d+\s*)?:\s*(-?\s*\d+\s*)?(:\s*(-?\s*\d+)?)?\s*$',
)


@marsh.schema.register
class SliceMarshalSchema(marsh.schema.MarshalSchema):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return isinstance(value, slice)
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`slice`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals a :class:`slice` into a mapping with '
            'the keys ``"start"``, ``"stop"`` and ``"step"``.'
        )

    def marshal(
        self,
    ) -> marsh.element.MappingElementType:
        return dict(
            start=self.value.start,
            stop=self.value.stop,
            step=self.value.step,
        )


@marsh.schema.register
class SliceUnmarshalSchema(marsh.schema.UnmarshalSchema[slice]):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return value is slice
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`slice`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Accepts input as strings in the python slicing format '
            '(i.e. "1:2", "::-1" e.t.c). '
            'Also accepts keys "start", "stop" and/or "step" or '
            'a sequence of up to 3 optional integer values. '
            'If a single integer is given, it is treated as "stop"'
        )

    def doc_field_type(
        self,
    ) -> str:
        return '<slice>'

    def doc_description(self) -> Optional[str]:
        return self.doc_static_description()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> slice:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            element = ()
        args: Tuple[Optional[int], ...]
        if marsh.utils.is_mapping(element):
            kwargs: Dict[str, Optional[int]] = marsh.cast(
                Dict[Literal['start', 'stop', 'step'], Optional[int]],  # type: ignore
                element,
            )
            args = tuple(kwargs.get(key, None) for key in ('start', 'stop', 'step'))
        elif marsh.utils.is_sequence(element):
            args = marsh.cast(Tuple[Optional[int], ...], element)  # type: ignore
            if len(args) > 3:
                raise marsh.errors.UnmarshalError(
                    f'{marsh.utils.get_type_name(self.value)}: at most '
                    '3 optional integers may be used to construct a '
                    f'slice. Got {element}',
                )
        else:
            element = str(element)
            if not slice_pattern.match(element):
                raise marsh.errors.UnmarshalError(
                    f'{marsh.utils.get_type_name(self.value)}: not valid '
                    f'slice syntax: "{element}"',
                )
            args = tuple(int(index) if index else None for index in element.split(':'))
        if not args:
            return slice(None)
        try:
            return slice(*args)
        except Exception:
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: failed '
                'to construct object',
            )



"""Collection of common errors used throughout the framework."""
import contextlib
import os
from typing import (
    Any,
    Callable,
    ClassVar,
    Final,
    Iterator,
    Type,
    TypeVar,
    overload,
)

import marsh


_E = TypeVar('_E', bound=BaseException)


_MARSH_FULL_ERROR_ENV: Final = 'MARSH_FULL_ERROR'


class MarshError(Exception):
    """Base class for marsh errors.

    Allows tracking of the path that produced the
    error.

    Arguments:
        *args: General exception arguments.
        path: The path that produced the error.
    """

    delimiter: ClassVar[str] = '.'

    def __init__(
        self,
        *args,
        path: str = '',
    ) -> None:
        super().__init__(*args)
        self.path = path

    def pretty(
        self,
        cause: bool = False,
    ) -> str:
        msg = self.__class__.__name__
        if self.args:
            msg += ': ' + ', '.join(map(str, self.args))
        if self.path:
            msg = f'{msg}\n\tpath: {self.path}'
        if cause:
            error: BaseException = self
            while error.__cause__ is not None:
                if not isinstance(error.__cause__, MarshError):
                    msg = f'{msg}\n\tcause: {error}'
                    break
                error = error.__cause__
        return msg

    def append(
        self,
        field: str,
    ) -> None:
        """Append a field to the path of this error."""
        self.path = marsh.path.append(
            path=self.path,
            field=field,
            delimiter=self.delimiter,
        )

    def prepend(
        self,
        field: str,
    ) -> None:
        """Prepend a field to the path of this error."""
        self.path = marsh.path.prepend(
            path=self.path,
            field=field,
            delimiter=self.delimiter,
        )


class PathError(MarshError, KeyError, IndexError):
    """Accessing non-existing fields or indices.

    Also used for path formatting errors.
    """


class UnmarshalError(MarshError):
    """Failure to unmarshal an element to a specific type."""
    pass


class MissingValueError(UnmarshalError):
    """Required value is missing."""
    pass


class MarshalError(MarshError):
    """Failure to serialize a value."""
    pass


class ConfigFileError(MarshError):
    """Base error for config files."""

    delimiter: ClassVar[str] = '/'


# TODO: Python 3.9: add [None] to return type.
@overload
def maybe_handle_error(
    callback: Callable[[MarshError], Any],
) -> contextlib.AbstractContextManager:
    ...


# TODO: Python 3.9: add [None] to return type.
@overload
def maybe_handle_error(
    callback: Callable[[_E], Any],
    error: Type[_E],
) -> contextlib.AbstractContextManager:
    ...


@contextlib.contextmanager
def maybe_handle_error(
    callback: Callable[[Any], Any],
    error: Type[BaseException] = MarshError,
) -> Iterator[None]:
    """Context manager that invokes a callback when catching an error.

    The error is propagated and the callback ignored if the environmental variable
    ``MARSH_FULL_ERROR`` is set and not empty.

    The error is passed to the callback.

    Arguments:
        callback: Callback function which may be called with the caught error.

    Returns:
        Context manager.
    """
    try:
        yield
    except error as e:
        if os.environ.get(_MARSH_FULL_ERROR_ENV, None):
            raise
        callback(e)


@contextlib.contextmanager
def prepend(
    field: Any,
) -> Iterator[None]:
    """Catches any marsh error, prepends a field to it and then re-raises the error.

    Arguments:
        field: The field to prepend to the marsh error path attribute.

    Returns:
        Context manager.
    """
    try:
        yield
    except MarshError as e:
        field = str(field)
        if field:
            e.prepend(field)
        raise



"""The baseline of types supported by the framework."""
from . import (
    annotated,
    any,
    bytes,
    complex,
    dataclass,
    datetime,
    defaultdict,
    enum,
    generic,
    literal,
    logger,
    mapping,
    namedtuple,
    namespace,
    none,
    omegaconf,
    primitive,
    protocol,
    sequence,
    set,
    slice,
    tuple,
    typeddict,
    typevar,
    union,
)


__all__ = (
    'annotated',
    'any',
    'bytes',
    'complex',
    'dataclass',
    'datetime',
    'defaultdict',
    'enum',
    'generic',
    'literal',
    'logger',
    'mapping',
    'namedtuple',
    'namespace',
    'none',
    'omegaconf',
    'primitive',
    'protocol',
    'sequence',
    'set',
    'slice',
    'tuple',
    'typeddict',
    'typevar',
    'union',
)



import collections
from typing import (
    Any,
    Callable,
    Mapping,
    TypeVar,
)

import marsh
from . import mapping


_T = TypeVar('_T')


class InitType:

    _sentinel: Any = object()

    def __init__(
        self,
        schema: marsh.schema.UnmarshalSchema[_T],
    ) -> None:
        self.schema = schema
        self.input: marsh.element.ElementType = self._sentinel

    def __call__(
        self,
    ) -> _T:
        if self.input is not self._sentinel:
            return self.schema.unmarshal(self.input)
        for self.input in (marsh.MISSING, 0, '', None):
            try:
                return self.schema.unmarshal(self.input)
            except Exception:
                pass
        self.input = self._sentinel
        raise marsh.errors.MarshError(
            'could not automatically initialize the type '
            f'{self.schema.doc_type()}',
        )


@marsh.schema.register(lower_priority=mapping.MappingUnmarshalSchema)
class DefaultDictUnmarshalSchema(
    mapping.MappingUnmarshalSchema[collections.defaultdict],
):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.initializer: Callable[[], Any]
        if marsh.utils.inspect_mapping_type(self.value).value is Any:
            self.initializer = lambda: None
        else:
            self.initializer = InitType(self.value_schema)

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_defaultdict_type(value)

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~collections.defaultdict`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals an input mapping into a defaultdict. '
            'If no default value is given in the type, all  '
            'values default to :data:`None`. An attempt is made '
            'to initialize the default value with a zero-value '
            '(zero, empty sequence, empty mapping, :data:`None` e.t.c). '
            'This is not always guaranteed to work if the default type '
            'requires specific arguments to be initialized. Use with '
            'caution.'
        )

    def doc_description(
        self,
    ) -> None:
        return None

    def construct(
        self,
        value: Mapping[Any, Any],
    ) -> collections.defaultdict:
        return collections.defaultdict(self.initializer, value)



from typing import (
    Any,
    Mapping,
    Optional,
    Sequence,
    get_args,
)


import marsh


@marsh.schema.register(priority=5)
class AnnotatedUnmarshalSchema(marsh.schema.UnmarshalSchema[Any]):

    schema: marsh.schema.UnmarshalSchema

    def __init__(
        self,
        value: Any,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(value, *args, **kwargs)
        self.schema = marsh.schema.UnmarshalSchema(
            get_args(self.value)[0],
            *args,
            **kwargs,
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return marsh.utils.is_annotated(value)
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':data:`~typing.Annotated`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unwraps and delegates unmarshaling '
            'to the underlying type.'
        )

    def __str__(
        self,
    ) -> str:
        return str(self.schema)

    def __repr__(
        self,
    ) -> str:
        return repr(self.schema)

    def select(
        self,
        path: str,
    ) -> marsh.schema.UnmarshalSchema:
        return self.schema.select(path)

    def doc(
        self,
        depth: int = 0,
    ) -> marsh.schema.UnmarshalSchema.Doc:
        return self.schema.doc(depth)

    def doc_type(
        self,
    ) -> str:
        return self.schema.doc_type()

    def doc_field_type(
        self,
    ) -> str:
        return self.schema.doc_field_type()

    def doc_default(
        self,
    ) -> Optional[str]:
        return self.schema.doc_default()

    def doc_description(
        self,
    ) -> Optional[str]:
        return self.schema.doc_description()

    def doc_fields(
        self,
        depth: int,
    ) -> Optional[Mapping[str, marsh.schema.UnmarshalSchema.Doc.Field]]:
        return self.schema.doc_fields(depth)

    def doc_special_fields(
        self,
    ) -> Optional[Sequence[marsh.schema.UnmarshalSchema.Doc.SpecialField]]:
        return self.schema.doc_special_fields()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> Any:
        value = self.schema.unmarshal(element)
        for ann in marsh.utils.get_annotations(self.value):
            try:
                if issubclass(ann, marsh.annotation.Annotation):
                    ann = ann()
            except TypeError:
                pass
            try:
                is_ann = isinstance(ann, marsh.annotation.Annotation)
            except TypeError:
                continue
            if is_ann:
                try:
                    value = ann(value)
                except marsh.errors.MarshError:
                    raise
                except Exception as e:
                    raise marsh.errors.UnmarshalError(
                        f'custom annotation {ann.__class__.__name__}'
                        f' failed: {e}',
                    )
        return value



from typing import Any

import marsh


@marsh.schema.register
class AnyUnmarshalSchema(marsh.schema.UnmarshalSchema[marsh.element.ElementType]):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return value is Any

    @staticmethod
    def doc_static_type() -> str:
        return ':data:`~typing.Any`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Any JSON-like data.\n'
            'Values should consist of one or '
            'more of the types :class:`int`, '
            ':class:`float`, :class:`str`, :class:`bool`, '
            ':class:`~collections.abc.Sequence` or '
            ':class:`~collections.abc.Mapping`. '
            'Can not contain missing values.'
        )

    def doc_type(
        self,
    ) -> str:
        return 'Any'

    def doc_field_type(
        self,
    ) -> str:
        return '<Any>'

    def doc_description(
        self,
    ) -> str:
        return self.doc_static_description()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> marsh.element.ElementType:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            raise marsh.errors.MissingValueError
        if marsh.element.has_missing(element):
            raise marsh.errors.MissingValueError(f'contains missing value: {element}')
        return element



from typing import (
    Sequence,
    TypeVar,
)

import marsh
from .. import UnmarshalSchema


_T = TypeVar('_T')


class UnionUnmarshalSchema(UnmarshalSchema[_T]):
    """A schema for a union of types.

    The order of the types is reflected in the
    order of attempted unmarshals. The value of
    the first type that is successfully unmarshaled
    is returned.

    If all types failed to unmarshal an error is raised."""

    schemas: Sequence[UnmarshalSchema]
    """The schemas for the union of types."""

    def __str__(
        self,
    ) -> str:
        return ' | '.join(map(str, self.schemas))

    def __repr__(
        self,
    ) -> str:
        return f'{super().__repr__()[:-1]}, schemas={tuple(map(repr, self.schemas))})'

    def doc_field_type(
        self,
    ) -> str:
        return ' | '.join(schema.doc_field_type() for schema in self.schemas)

    def select(
        self,
        path: str,
    ) -> UnmarshalSchema:
        if (
            marsh.utils.is_optional(self.value)
            and len(self.schemas) == 2
        ):
            for schema in self.schemas:
                if schema.value not in (None, type(None)):
                    return schema.select(path)
        if not path:
            return self
        raise marsh.errors.PathError(
            f'{marsh.utils.get_type_name(self.value)}: ambiguous '
            'traversal route (can not traverse union nodes)',
        )

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _T:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
        for schema in self.schemas:
            try:
                return schema.unmarshal(element)
            except Exception:
                pass
        if marsh.utils.is_missing(element):
            raise marsh.errors.MissingValueError
        raise marsh.errors.UnmarshalError(
            f'{marsh.utils.get_type_name(self.value)}: '
            f'failed to unmarshal element: {element}',
        )



"""Annotations that can be used with :data:`typing.Annotated`.

Annotations may be used to customize the unmarshaling behavior
through transforms or validations.

.. code-block:: python

    from typing import Annotated
    import marsh

    UnsignedInt = Annotated[int, marsh.annotation.Unsigned]
    value = marsh.unmarshal(UnsignedInt, 5)  # works
    value = marsh.unmarshal(UnsignedInt, -5)  # fails


Create your own annotations by inheriting :class:`~marsh.annotation.Annotation`.

.. code-block:: python

    from typing import Annotated
    import marsh

    class Abs(marsh.annotation.Annotation):

        def __call__(
            self,
            value
        ):
            return abs(value)

    AbsInt = Annotated[int, Abs]
    assert marsh.unmarshal(AbsInt, -5) == 5
"""
import numbers
from typing import (
    Any,
    Sized,
    TypeVar,
)


_T = TypeVar('_T', bound=Any)
_NumberType = TypeVar('_NumberType', bound=numbers.Number)
_SizedType = TypeVar('_SizedType', bound=Sized)


class Annotation:
    """Base class for annotations recognized by :mod:`marsh`.

    These can be used with :data:`typing.Annotated` to include
    validation or transformation of the unmarshal result."""

    def __call__(
        self,
        value: _T,
    ) -> _T:
        """Validate and/or transform the input value.

        Arguments:
            value: The input value.

        Returns:
            The validated and/or transformed value.
        """
        return value


class Positive(Annotation):
    """Validate that a number is non-zero and positive."""

    def __call__(
        self,
        value: _NumberType,
    ) -> _NumberType:
        if value <= 0:  # type: ignore
            raise ValueError(
                f'positive value required, got {value}',
            )
        return value


class Negative(Annotation):
    """Validate that a number is non-zero and negative."""

    def __call__(
        self,
        value: _NumberType,
    ) -> _NumberType:
        if value >= 0:  # type: ignore
            raise ValueError(
                f'negative value required, got {value}',
            )
        return value


class Unsigned(Annotation):
    """Validate that a number is zero or higher."""

    def __call__(
        self,
        value: _NumberType,
    ) -> _NumberType:
        if value < 0:  # type: ignore
            raise ValueError(
                f'unsigned value required, got {value}',
            )
        return value


class Populated(Annotation):
    """Validate that a collection contains at least one item."""

    def __call__(
        self,
        value: _SizedType,
    ) -> _SizedType:
        if len(value) == 0:
            raise ValueError(
                f'value must have size larger than 0: {value}',
            )
        return value



from typing import (
    Any,
    get_args,
    get_origin,
)

import marsh
from . import sequence


@marsh.schema.register
class SetUnmarshalSchema(sequence.SequenceUnmarshalSchema):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        type_ = Any
        if (_args := get_args(self.value)):
            type_ = _args[0]
        self.value_schema = marsh.schema.UnmarshalSchema(type_)

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return (
                get_origin(value) is set
                or issubclass(value, set)
            )
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`set`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals a sequence input into '
            'a set. Any duplicate values will be '
            'discarded.'
        )



import inspect
import types
from typing import (
    Any,
    ForwardRef,
    TypeVar,
    get_type_hints,
)

import marsh
from . import structured


_T = TypeVar('_T')


class CallableUnmarshalSchema(structured.StructuredUnmarshalSchema[_T]):
    """Unmarshals the arguments to a callable and uses them to call it."""

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        try:
            sig = inspect.signature(self.value)
        except Exception:
            raise ValueError(
                f'failed to retrieve callable signature: {self.value}',
            )
        try:
            if inspect.isclass(self.value):
                annotations = get_type_hints(self.value.__init__)
            else:
                annotations = get_type_hints(self.value)
        except Exception:
            annotations = {}
        positional = {}
        keyword = {}
        var_positional = None
        var_keyword = None
        for name, param in sig.parameters.items():
            if (
                param.annotation is param.empty
                or isinstance(param.annotation, (str, ForwardRef))
            ):
                type_ = annotations.get(name, Any)
            else:
                type_ = param.annotation
            default = marsh.MISSING
            if param.default is not param.empty:
                default = param.default
            with marsh.errors.prepend(name):
                schema = marsh.schema.UnmarshalSchema(type_, default=default)
            if param.kind == param.POSITIONAL_ONLY:
                positional[name] = schema
            elif param.kind == param.POSITIONAL_OR_KEYWORD:
                keyword[name] = schema
            elif param.kind == param.KEYWORD_ONLY:
                keyword[name] = schema
            elif param.kind == param.VAR_POSITIONAL:
                var_positional = (name, schema)
            elif param.kind == param.VAR_KEYWORD:
                var_keyword = (name, schema)
        self.positional = tuple(positional)
        schemas = {}
        schemas.update(positional)
        if var_positional:
            name, schema = var_positional
            schemas[f'*{name}'] = schema
        schemas.update(keyword)
        if var_keyword:
            name, schema = var_keyword
            schemas[f'**{name}'] = schema
        self.schemas = types.MappingProxyType(schemas)



from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
)

import marsh
from .. import UnmarshalSchema


_T = TypeVar('_T')


class StructuredUnmarshalSchema(UnmarshalSchema[_T]):
    """A schema for types with fixed attributes
    that are recursively unmarshaled.

    Supports both positional arguments and keyword
    arguments. Variable versions of these arguments
    (typically ``*args``/``**kwargs``) are also supported
    and are marked by ``*`` or ``**`` infront of their
    names in the :attr:`schemas` attribute of this class."""

    schemas: Mapping[str, UnmarshalSchema]
    """The schemas for the fixed attributes."""

    positional: Sequence[str] = ()
    """Names of attributes that can only be used as positional arguments."""

    def __str__(
        self,
    ) -> str:
        return (
            f'{marsh.utils.get_type_name(self.value)}('
            + ', '.join(
                f'{name}: {schema}'
                for name, schema
                in self.schemas.items()
            )
            + ')'
        )

    def __repr__(
        self,
    ) -> str:
        return (
            f'{super().__repr__()[:-1]}, '
            f'schemas={repr(self.schemas)})'
        )

    def validate_defaults(
        self,
        args: Optional[Sequence] = None,
        kwargs: Optional[Mapping[str, Any]] = None,
    ) -> None:
        if args and kwargs:
            for i, name in enumerate(self.schemas):
                if i < len(args) and name in kwargs:
                    raise TypeError(
                        'received the same argument twice '
                        'through `default_args` and '
                        f'`default_kwargs`: {name}',
                    )

    def select(
        self,
        path: str,
    ) -> UnmarshalSchema:
        if not path:
            return self
        field, path = marsh.path.head(path)
        if not field:
            if not path:
                return self
            return self.select(path)
        for name, schema in self.schemas.items():
            if field.strip('*') == name.strip('*'):
                with marsh.errors.prepend(field):
                    return schema.select(path)
        raise marsh.errors.PathError(
            marsh.utils.get_closest_error_message(
                value=field,
                candidates=self.schemas,
                key='key',
            ),
        )

    def doc(
        self,
        depth: int = 0,
        default_args: Optional[Sequence] = None,
        default_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> UnmarshalSchema.Doc:
        return self.Doc(
            type=self.doc_type(),
            default=self.doc_default(),
            description=self.doc_description(),
            fields=self.doc_fields(
                max(0, depth),
                default_args=default_args,
                default_kwargs=default_kwargs,
            ),
            special_fields=self.doc_special_fields(),
        )

    def doc_field_type(
        self,
    ) -> str:
        return f'{marsh.utils.get_type_name(self.value)}(...)'

    def doc_fields(
        self,
        depth: int,
        default_args: Optional[Sequence] = None,
        default_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Optional[Mapping[str, UnmarshalSchema.Doc.Field]]:
        if depth <= 0:
            return None
        self.validate_defaults(
            args=default_args,
            kwargs=default_kwargs,
        )
        default_args = list(default_args or ())
        default_kwargs = dict(default_kwargs or {})
        fields: Dict[str, UnmarshalSchema.Doc.Field] = {}
        for i, (name, schema) in enumerate(self.schemas.items()):
            fields[name] = UnmarshalSchema.Doc.Field(
                doc=schema.doc(depth - 1),
                type=schema.doc_field_type(),
                description=marsh.utils.get_attribute_description(
                    self.value,
                    name.strip('*'),
                ),
            )
            default = None
            if name.startswith('**'):
                if default_kwargs:
                    default = str(default_kwargs)
                    default_kwargs = {}
            elif name.startswith('*'):
                if default_args:
                    default = str(default_args)
                    default_args = []
            if i < len(default_args):
                default = str(default_args.pop(0))
            if name in default_kwargs:
                default = str(default_kwargs.pop(name))
            if default:
                fields[name].doc.default = default
        return fields

    def unmarshal(
        self,
        element: marsh.element.ElementType,
        default_args: Optional[Sequence] = None,
        default_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> _T:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            element = {}
        element_list: List[Any] = []
        element_dict: Dict[str, Any] = {}
        args: List = []
        kwargs: Dict[str, Any] = {}
        if marsh.utils.is_mapping(element):
            element_dict = dict(element)
        elif marsh.utils.is_sequence(element):
            element_list = list(element)
        else:
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: expected a '
                f'mapping or sequence element, got: {element}',
            )
        self.validate_defaults(
            args=default_args,
            kwargs=default_kwargs,
        )
        default_args = list(default_args or ())
        default_kwargs = dict(default_kwargs or {})
        for name, schema in self.schemas.items():
            with marsh.errors.prepend(name.strip('*')):
                if name.startswith('**'):
                    for key in tuple(default_kwargs):
                        value = default_kwargs.pop(key)
                        if key not in element_dict:
                            kwargs[key] = value
                    for key in tuple(element_dict):
                        with marsh.errors.prepend(key):
                            kwargs[key] = schema.unmarshal(element_dict.pop(key))
                elif name.startswith('*'):
                    for i in range(len(element_list)):
                        with marsh.errors.prepend(i):
                            args.append(schema.unmarshal(element_list.pop(0)))
                    args.extend(default_args)
                    default_args = []
                else:
                    if element_list:
                        args.append(schema.unmarshal(element_list.pop(0)))
                        default_args and default_args.pop(0)
                    else:
                        if default_args:
                            value = default_args.pop(0)
                            if name not in element_dict:
                                kwargs[name] = value
                        if name in default_kwargs:
                            value = default_kwargs.pop(name)
                            if name not in element_dict:
                                kwargs[name] = value
                        if name not in kwargs:
                            kwargs[name] = schema.unmarshal(
                                element_dict.pop(
                                    name,
                                    marsh.MISSING,
                                ),
                            )
                        if name in self.positional:
                            args.append(kwargs.pop(name))
        for key in element_dict:
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: '
                + marsh.utils.get_closest_error_message(
                    key,
                    self.schemas,
                    key='key',
                ),
            )
        if element_list:
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: received too '
                f'many arguments. unconsumed arguments: {element_list}',
            )
        try:
            return self.construct(*args, **kwargs)
        except marsh.errors.MarshError:
            raise
        except Exception:
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: failed to '
                f'construct object using *args={args}, **kwargs={kwargs}',
            )

    def construct(
        self,
        *args,
        **kwargs,
    ) -> _T:
        """Takes the unmarshaled arguments to the type as input and
        initializes an instance of the type using these arguments.

        Returns:
            The initialized value.
        """
        return self.value(*args, **kwargs)



from typing import (
    Any,
    Type,
    TypeVar,
    Union,
)

import marsh


_P = TypeVar('_P', int, float, bool, str)
Primitive = Union[int, float, bool, str]
PrimitiveType = Union[Type[int], Type[float], Type[bool], Type[str]]


@marsh.schema.register
class PrimitiveMarshalSchema(marsh.schema.MarshalSchema):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_primitive(value)

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`int`, :class:`float`, :class:`bool`, :class:`str`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Marshals any instance of a primitive class or subclass '
            'into its baseclass value.'
        )

    def marshal(
        self,
    ) -> Primitive:
        if isinstance(self.value, bool):
            return bool(self.value)
        elif isinstance(self.value, int):
            return int(self.value)
        elif isinstance(self.value, float):
            return float(self.value)
        else:
            return str(self.value)


@marsh.schema.register
class PrimitiveUnmarshalSchema(marsh.schema.UnmarshalSchema[_P]):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return (
            marsh.utils.is_primitive_type(value)
            or marsh.utils.is_literal_string(value)
        )

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`int`, :class:`float`, :class:`bool`, :class:`str`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals any primitive class or its subclass. '
            'Special rules for :class:`bool` is documented in '
            ':func:`~marsh.utils.primitive_to_bool`.'
        )

    def get_type(
        self,
    ) -> Type[_P]:
        if marsh.utils.is_literal_string(self.value):
            return str  # type: ignore
        return self.value

    def doc_field_type(
        self,
    ) -> str:
        return f'<{marsh.utils.get_type_name(self.value)}>'

    def doc_description(
        self,
    ) -> None:
        return None

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _P:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            raise marsh.errors.MissingValueError
        if not marsh.utils.is_primitive(element):
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: expected '
                f'a primitive value, got: {element}',
            )
        try:
            return marsh.utils.cast_primitive(
                self.get_type(),
                element,
            )
        except Exception:
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: '
                f'could not convert: {element}',
            )



import textwrap
from typing import Optional

import marsh
from . import common


def format_terminal_value(
    value: str,
    width: Optional[int] = None,
    description: Optional[str] = None,
    indent: int = 2,
    description_indent: int = 24,
    description_margin: int = 4,
) -> str:
    """Create a help message in the same style as
    :class:`argparse.ArgumentParser` for a value and
    its description

    Arguments:
        value: The string value.
        width: The width of the message.
        description: A text description for the value.
        indent: The indentation of the value
            in number of blank spaces.
        description_indent: The indentation of the
            description in number of blank spaces.
        description_margin: The minimum number of
            blank spaces between the description
            and the value. The description will be
            started on the next line if the resulting
            margin would be smaller than this on the
            line containing the argument.

    Returns:
        The help message.
    """
    if width is None:
        width = marsh.utils.get_terminal_width()
    elif width <= 0:
        raise ValueError('`width` must be a positive integer')
    value = f'{" " * indent}{value}'
    value = '\n'.join(textwrap.wrap(value, width=width))
    if not description:
        return value
    fill_indent = ' ' * description_indent
    fill_init_indent = fill_indent
    value_len = len(value.split('\n')[-1])
    if description_indent - value_len >= description_margin:
        fill_init_indent = fill_indent[value_len:]
    else:
        value = f'{value}\n'
    description = textwrap.fill(
        text=description,
        width=width,
        initial_indent=fill_init_indent,
        subsequent_indent=fill_indent,
    )
    return f'{value}{description}'


def format_unmarshal_fields(
    source: common.UnmarshalDocType,
    prefix_path: str = '',
    width: Optional[int] = None,
    depth: int = 1,
    indent: int = 2,
    depth_indent_factor: int = 2,
    description_indent: int = 24,
    description_margin: int = 4,
) -> str:
    if width is None:
        width = marsh.utils.get_terminal_width()
    description_indent = min(description_indent, width - 1)

    def format_fields(
        doc: marsh.schema.UnmarshalSchema.Doc,
        path: str = '',
        current_depth: int = 0,
    ) -> str:
        if current_depth >= depth:
            return ''
        fields = []
        for name, field in (doc.fields or {}).items():
            name = marsh.path.append(path, name)
            value = name
            if field.type:
                value = f'{value}: {field.type}'
            if field.doc.default:
                value = f'{value} = {field.doc.default}'
            fields.append(
                format_terminal_value(
                    value=value,
                    width=width,
                    description=field.description,
                    indent=indent + current_depth * depth_indent_factor,
                    description_indent=description_indent,
                    description_margin=description_margin,
                ),
            )
            fields.append(
                format_fields(
                    doc=field.doc,
                    path=name,
                    current_depth=current_depth + 1,
                ),
            )
        for special_field in (doc.special_fields or ()):
            fields.append(
                format_terminal_value(
                    value=special_field.value,
                    width=width,
                    description=special_field.description,
                    indent=indent + current_depth * depth_indent_factor,
                    description_indent=description_indent,
                    description_margin=description_margin,
                ),
            )
        return '\n\n'.join(filter(None, fields))
    return format_fields(
        doc=common.get_unmarshal_doc(
            source=source,
            depth=depth,
        ),
        path=prefix_path,
    )


def format_unmarshal_help(
    source: common.UnmarshalDocType,
    prefix_path: str = '',
    width: Optional[int] = None,
    depth: int = 1,
    field_indent: int = 2,
    depth_indent_factor: int = 2,
    description_indent: int = 24,
    description_margin: int = 4,
    omit_type: bool = False,
    omit_description: bool = False,
) -> str:
    help_msg = ''
    doc = common.get_unmarshal_doc(
        source=source,
        depth=depth,
    )
    if doc.type and not omit_type:
        help_msg = doc.type
    if doc.description and not omit_description:
        if help_msg:
            help_msg = f'{help_msg}\n\n{doc.description}'
        else:
            help_msg = doc.description
    fields_msg = format_unmarshal_fields(
        source=doc,
        prefix_path=prefix_path,
        width=width,
        depth=depth,
        indent=field_indent,
        depth_indent_factor=depth_indent_factor,
        description_indent=description_indent,
        description_margin=description_margin,
    )
    if fields_msg:
        if help_msg:
            help_msg = f'{help_msg}\n\n{fields_msg}'
        else:
            help_msg = fields_msg
    return help_msg



from typing import (
    Iterable,
    Literal,
    Optional,
    Sequence,
)
from . import common


INDENT = ' ' * 4


def format_table(
    headers: Sequence,
    rows: Iterable[Sequence],
    title: Optional[str] = None,
    widths: Optional[Sequence] = None,
) -> str:
    """Create a RST list-table.

    Any newlines in the headers or rows are
    replaced with whitespace.

    Arguments:
        headers: The header names.
        rows: Iterable of rows, each
            row containing the same number
            of values as there are headers.
        title: A title for the table.
        widths: The width of columns. If given,
            must contain the same number of values
            as there are headers.

    Returns:
        The table in RST format.
    """
    title = (title or '') and f' {title}'
    table = (
        f'.. list-table::{title}\n'
        f'{INDENT}:header-rows: 1\n'
    )
    if widths:
        table += f'{INDENT}:widths: {", ".join(map(str, widths))}\n'
    table += '\n'
    table_rows = []
    for row in (headers,) + tuple(rows):
        for i, col in enumerate(row):
            if i:
                indent = f'{INDENT}  - '
            else:
                indent = f'{INDENT}* - '
            table_rows.append(f'{indent}{col}'.replace('\n', ' '))
    return table + '\n'.join(table_rows) + '\n'


def format_types(
    source: common.RegisteredSchemasType,
    section: Literal['#', '*', '=', '-', '^', '"'] = '^',
) -> str:
    """Return a RST-formatted string containing
    registered schema types that have static documention.

    A description below each type is included when available.

    Arguments:
        source: The source of the schemas.
        section: The section character.

    Returns:
        The formatted RST.
    """
    sections = []
    for schema in common.get_registered_schemas(source):
        type_ = schema.doc_static_type()
        if not type_:
            continue
        underline = section * len(type_)
        overline = ''
        if section in '#*':
            overline = underline + '\n'
        sections.append(
            f'{overline}{type_}\n{underline}\n\n'
            f'{schema.doc_static_description() or ""}',
        )
    return '\n\n'.join(sections)


def format_types_table(
    source: common.RegisteredSchemasType,
    title: Optional[str] = None,
) -> str:
    """Return a RST-formatted table containing
    registered schema types that have static documention.

    A description for each type is included when available.

    Arguments:
        source: The source of the schemas.

    Returns:
        The formatted RST table.
    """
    return format_table(
        headers=('Type', 'Description'),
        widths=(30, 50),
        rows=(
            (
                schema.doc_static_type(),
                schema.doc_static_description() or '',
            )
            for schema
            in common.get_registered_schemas(source)
            if schema.doc_static_type()
        ),
        title=title,
    )



"""Tools for parsing strings into elements and/or commands."""
import contextlib
import dataclasses
from typing import (
    Any,
    Dict,
    Final,
    Iterator,
    List,
    Mapping,
    MutableSequence,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
)

import marsh


SET_VALUE_CHARACTERS: Final = '+='
SET_CONFIG_CHARACTERS: Final = '+@'
DELETE_VALUE_CHARACTERS: Final = '~'
OPENING_SEQUENCE_CHARACTERS: Final = '(['
OPENING_MAPPING_CHARACTERS: Final = '{'
QUOTE_CHARACTERS: Final = '\'"'
ESCAPE: Final = '\\'


_MaybeMutableSource = Union[str, MutableSequence[str]]


class _ParseFunction(Protocol):

    def __call__(
        self,
        source: Union[str, MutableSequence[str]],
        terminal_chars: str = '',
    ) -> 'marsh.element.ElementType':
        ...


class ParseError(Exception):

    def __init__(
        self,
        *args,
        remaining: _MaybeMutableSource,
        source: Optional[_MaybeMutableSource] = None,
    ) -> None:
        super().__init__(*args)
        if not isinstance(remaining, str):
            remaining = ''.join(remaining)
        self.remaining = remaining
        if source and not isinstance(source, str):
            source = ''.join(source)
        self.source = source

    def __str__(
        self,
    ) -> str:
        rep = super().__str__()
        if not self.remaining:
            return rep
        if not self.source:
            return f'{rep}\n... {self.remaining[:70]}\n{" " * 3}^'
        index = min(len(self.source) - 1, max(0, len(self.source) - len(self.remaining)))
        source = self.source
        if index > 35:
            source = f'... {source[index - 35:]}'
            index = index - 35 + 4
        if len(source) > 70:
            source = f'{source[:70]} ...'
        return f'{rep}\n{source}\n{" " * index}^'


@contextlib.contextmanager
def parsing(
    source: _MaybeMutableSource,
    terminal_chars: str = '',
) -> Iterator[MutableSequence[str]]:
    if isinstance(source, str):
        original_source = source
        source = list(source)
    else:
        original_source = ''.join(source)
    try:
        yield source
    except ParseError as e:
        e.source = original_source
        raise e
    remaining = list(source)
    consume_whitespace(remaining)
    if remaining and peek(remaining) not in terminal_chars:
        terminals: Union[str, Tuple[str, ...]]
        terminals = terminal_chars
        if len(terminals) > 1:
            terminals = tuple(terminals)
        raise ParseError(
            (
                'parsing did not terminate correctly at ' + (
                    f'terminal character: {terminals}'
                    if terminals else
                    'end of input string'
                )
            ),
            remaining=remaining,
            source=original_source,
        )


def is_whitespace(
    source: Sequence[str],
) -> bool:
    if not isinstance(source, str):
        source = ''.join(source)
    return source.isspace()


def consume_whitespace(
    source: MutableSequence[str],
) -> None:
    while source and source[0].isspace():
        pop(source)


def peek(
    source: Sequence[str],
) -> str:
    if source:
        return source[0]
    return ''


def pop(
    source: MutableSequence[str],
) -> str:
    if not source:
        raise ValueError('can not pop from empty source')
    return source.pop(0)


@dataclasses.dataclass
class Override:

    path: str

    def apply(
        self,
        element: marsh.element.ElementType,
        **kwargs,
    ) -> marsh.element.ElementType:
        raise NotImplementedError


@dataclasses.dataclass
class SetValue(Override):

    value: marsh.element.ElementType

    combine: bool

    def apply(
        self,
        element: marsh.element.ElementType,
        **kwargs,
    ) -> marsh.element.ElementType:
        return marsh.element.override(
            element=element,
            path=self.path,
            value=self.value,
            combine=self.combine,
        )


@dataclasses.dataclass
class SetConfig(Override):

    names: Sequence[str]

    combine: bool

    def apply(
        self,
        element: marsh.element.ElementType,
        root: Optional[str] = None,
        keep_meta: bool = False,
        **kwargs,
    ) -> marsh.element.ElementType:
        return marsh.element.override(
            element=element,
            path=self.path,
            value=marsh.config.load(
                *self.names,
                root=root,
                keep_meta=keep_meta,
                concatenate=True,
            ),
            combine=self.combine,
        )


@dataclasses.dataclass
class DeleteValue(Override):

    def apply(
        self,
        element: marsh.element.ElementType,
        **kwargs,
    ) -> marsh.element.ElementType:
        if not (
            marsh.utils.is_mapping(element)
            or marsh.utils.is_sequence(element)
        ):
            raise marsh.errors.PathError(
                'can not remove value for '
                f'non-existing path: {self.path}',
            )
        return marsh.element.remove(
            element=element,
            path=self.path,
        )


def override(
    source: Union[str, MutableSequence[str]],
) -> Override:
    """Parse an override.
    Accepts the format
    ``path.to.field@name1,name2``/``path.to.field+@name1,name2``
    for config assignments and
    ``path.to.field=value``/``path.to.field+=value``/
    ``~path.to.field`` for value assignments/deletion.

    Arguments:
        source: The string source to parse

    Returns:
        The override
    """
    with parsing(source) as source:
        consume_whitespace(source)
        delete = False
        if peek(source) == '~':
            delete = True
            pop(source)
        path = string(
            source,
            terminal_chars=SET_VALUE_CHARACTERS + SET_CONFIG_CHARACTERS,
        )
        if delete:
            consume_whitespace(source)
            if source:
                raise ParseError(
                    f'unexpected reserved character: "{peek(source)}". '
                    'The delete override should be in the form of `~path.to.field`',
                    remaining=source,
                )
            return DeleteValue(path=path)
        if peek(source) not in (SET_VALUE_CHARACTERS + SET_CONFIG_CHARACTERS):
            raise ParseError(
                'expected `@`, `+@`, `=`, `+=` or `~` in override',
                remaining=source,
            )
        missing_err = 'expected `@`, `+@`, `=`, `+=` or `~` in override'
        combine = False
        if peek(source) == '+':
            pop(source)
            combine = True
            if peek(source) not in '=@':
                missing_err = 'expected `+@`, or `+=`'
        if peek(source) == '@':
            pop(source)
            return SetConfig(
                path=path,
                names=config_names(source),
                combine=combine,
            )
        if peek(source) == '=':
            pop(source)
            return SetValue(
                path=path,
                value=element(source),
                combine=combine,
            )
        raise ParseError(
            missing_err,
            remaining=source,
        )


def config_names(
    source: Union[str, MutableSequence[str]],
) -> Sequence[str]:
    """Parse the names of a config assignment as a
    comma-separated list of strings. List bracket/parenthesis
    `()`/`[]` may be omitted.

    Parsing terminates at end of input source.
    String is split at `,` unless escaped or within a pair of quotation
    marks.
    Quotation marks and escape characters are not kept (unless escaped).

    Arguments:
        source: The string source to parse.

    Returns:
        The names.
    """
    with parsing(source) as source:
        consume_whitespace(source)
        if peek(source) in '[(':
            return tuple(map(str, sequence(source, element_parser=string)))
        names: List[str] = []
        while source:
            if names:
                if peek(source) != ',':
                    raise ParseError(
                        'expected `,` or the termination of the input string',
                        remaining=source,
                    )
                pop(source)
            names.append(string(source, terminal_chars=','))
            consume_whitespace(source)
        return names


def string(
    source: Union[str, MutableSequence[str]],
    terminal_chars: str = '',
) -> str:
    """Parse a single string.

    Reserved characters are allowed by being
    escaped (`\\`) or for the string (or part of the
    string) to be enclosed in quotes (``"`` or ``'``).

    Quotes are discarded before returning the result.

    Arguments:
        source: The string source to parse
        terminal_chars: Any characters that should terminate the parsing process (
            unless escaped or within a set of quotes). If none are given the
            entire input is consumed.

    Returns:
        The parsed string.
    """
    with parsing(source, terminal_chars) as source:
        consume_whitespace(source)
        if not source:
            return ''
        quote: Optional[str] = None
        chars: List[str] = []
        if peek(source) in QUOTE_CHARACTERS:
            quote = pop(source)
        while (char := peek(source)) not in terminal_chars or quote:
            if char == ESCAPE:
                if len(source) < 2:
                    raise ParseError(
                        'a character must succeed the escape character `\\`',
                        remaining=source,
                    )
                pop(source)
                chars.append(pop(source))
                continue
            if char == quote:
                pop(source)
                quote = None
                break
            chars.append(pop(source))
        if quote:
            raise ParseError(
                f'expected closing quote `{quote}`',
                remaining=source,
            )
        return ''.join(chars)


def terminal(
    source: Union[str, MutableSequence[str]],
    terminal_chars: str = '',
) -> Union[None, int, float, bool, str]:
    """Parse a terminal value.

    None | int | float | string | bool

    Arguments:
        source: The string source to parse
        terminal_chars: Any characters that should terminate the parsing process (
            unless escaped or within a set of quotes). If none are given the
            entire input is consumed.

    Returns:
        The parsed value.
    """
    if isinstance(source, str):
        original_source = source
        source = list(source)
    else:
        original_source = ''.join(source)
    consume_whitespace(source)
    if not source:
        return ''
    quoted = peek(source) in QUOTE_CHARACTERS
    try:
        value = string(
            source,
            terminal_chars=terminal_chars,
        )
    except ParseError as e:
        e.source = original_source
        raise e
    if quoted or not value:
        return value
    try:
        fl_value = float(value)
        if fl_value.is_integer():
            return int(fl_value)
        return fl_value
    except ValueError:
        pass
    if value.lower() == 'true':
        return True
    if value.lower() == 'false':
        return False
    if value.lower() == 'null':
        return None
    return value


def element(
    source: Union[str, MutableSequence[str]],
    terminal_chars: str = '',
) -> 'marsh.element.ElementType':
    """Parse a single element.

    Returns a mapping, sequence or terminal element.

    See :func:`mapping`, :func:`sequence` and :func:`terminal`
    for more information parsing rules.

    Arguments:
        source: The string source to parse
        terminal_chars: Expected characters that can succeed the part
            of the input string that contains the element. If not given,
            an exception is raised if there is anything other than
            whitespace after the element.

    Returns:
        The parsed element.
    """
    with parsing(source, terminal_chars) as source:
        consume_whitespace(source)
        if not source:
            return ''
        if peek(source) == '{':
            return mapping(
                source,
                terminal_chars=terminal_chars,
            )
        elif peek(source) in '([':
            return sequence(
                source,
                terminal_chars=terminal_chars,
            )
        return terminal(
            source,
            terminal_chars=terminal_chars,
        )


def sequence(
    source: Union[str, MutableSequence[str]],
    terminal_chars: str = '',
    element_parser: _ParseFunction = element,
) -> Sequence:
    """Parse a single sequence of elements.

    See :func:`element` for defenition of an element.

    Elements should be separated by ``,``.
    The sequence should be enclosed in ``()`` or ``[]``.

    Arguments:
        source: The string source to parse
        terminal_chars: Expected characters that can succeed the part
            of the input string that contains the sequence. If not given,
            an exception is raised if there is anything other than whitespace
            after the sequence.
        element_parser: Each element of the sequence will be parsed using this parser.

    Returns:
        Sequence of elements.
    """
    with parsing(source, terminal_chars) as source:
        consume_whitespace(source)
        char = peek(source)
        if not (char and char in '(['):
            raise ParseError(
                'expected a sequence opening bracket `[` or parenthesis `(`',
                remaining=source,
            )
        closing_char = ']' if pop(source) == '[' else ')'
        consume_whitespace(source)
        value: List[marsh.element.ElementType] = []
        while peek(source) != closing_char:
            consume_whitespace(source)
            if value:
                if peek(source) != ',':
                    raise ParseError(
                        'expected `,` in sequence',
                        remaining=source,
                    )
                pop(source)
            consume_whitespace(source)
            value.append(
                element_parser(
                    source,
                    terminal_chars=',' + closing_char,
                ),
            )
        consume_whitespace(source)
        if peek(source) != closing_char:
            raise ParseError(
                f'expected `{closing_char}` as terminal character for sequence',
                remaining=source,
            )
        pop(source)
        return tuple(value)


def mapping(
    source: Union[str, MutableSequence[str]],
    terminal_chars: str = '',
    key_parser: _ParseFunction = string,
    value_parser: _ParseFunction = element,
) -> Mapping:
    """Parse a single mapping of strings to elements.

    See :func:`element` for defenition of an element.

    Keys and values should be separated by ``:``
    Key-value pairs should be separated by ``,``.
    The mapping should be enclosed in ``{}``.

    Arguments:
        source: The string source to parse
        terminal_chars: Expected characters that can succeed the part
            of the input string that contains the mapping. If not given,
            an exception is raised if there is anything other than whitespace
            after the mapping.
        key_parser: Each key of the mapping will be parsed using this parser.
        value_parser: Each value of the mapping will be parsed using this parser.

    Returns:
        The mapping of strings to elements.
    """
    with parsing(source, terminal_chars) as source:
        consume_whitespace(source)
        if peek(source) != '{':
            raise ParseError(
                'expected a mapping opening bracer `{`',
                remaining=source,
            )
        pop(source)
        consume_whitespace(source)
        value: Dict[Any, Any] = {}
        while peek(source) not in '}':
            if value:
                if peek(source) != ',':
                    raise ParseError(
                        'expected `,` in mapping',
                        remaining=source,
                    )
                pop(source)
            consume_whitespace(source)
            key = key_parser(
                source,
                terminal_chars=':',
            )
            consume_whitespace(source)
            if peek(source) != ':':
                raise ParseError(
                    (
                        'expected `:` in mapping' + (
                            f' after key "{key}"'
                            if key else ''
                        )
                    ),
                    remaining=source,
                )
            pop(source)
            consume_whitespace(source)
            value[key] = value_parser(
                source,
                terminal_chars=',}',
            )
            consume_whitespace(source)
        if peek(source) != '}':
            raise ParseError(
                'expected `}` as terminal character for mapping',
                remaining=source,
            )
        pop(source)
        return value



from typing import (
    Callable,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import TypeAlias

import marsh
from . import core


_S = TypeVar('_S', bound=core.Schema)
_RelativePriority: TypeAlias = \
    marsh.utils.PriorityOrder.RelativePriority[Type[core.Schema]]


@overload
def register(
    schema: Type[_S],
    /,
    *,
    priority: int = 0,
    lower_priority: _RelativePriority = None,
    higher_priority: _RelativePriority = None,
    replace: bool = False,
) -> Type[_S]:
    ...


@overload
def register(
    *,
    priority: int = 0,
    lower_priority: _RelativePriority = None,
    higher_priority: _RelativePriority = None,
    replace: bool = False,
) -> Callable[[Type[_S]], Type[_S]]:
    ...


def register(
    schema: Optional[Type[_S]] = None,
    /,
    *,
    priority: int = 0,
    lower_priority: _RelativePriority = None,
    higher_priority: _RelativePriority = None,
    replace: bool = False,
) -> Union[Callable[[Type[_S]], Type[_S]], Type[_S]]:
    """Register a new schema (decorator).

    Relative priorities can be given as ``higher`` and ``lower``,
    where any schema in ``higher`` means that the schema being registered
    should have a lower priority than that, the opposite is true for ``lower``.

    The relative priority (if given) takes precedence over the base priority.

    Arguments:
        schema: The schema to register.
        priority: Base priority level.
        lower_priority: Any other schemas that should
            have lower priority compared to the new schema.
        higher_priority: Any other schemas that should
            have higher priority compared to the new schema.
        replace: Allow replacing existing registered schema
            with same name.

    Returns:
        Decorator if ``schema`` was not given, else ``schema``.
    """
    def _register(
        schema,
    ):
        schema.registry.register(
            schema,
            priority=priority,
            lower_priority=lower_priority,
            higher_priority=higher_priority,
            replace=replace,
        )
        return schema
    if schema is None:
        return _register
    return _register(schema)



"""Mapping schema support for subclasses of :class:`collections.abc.Mapping`."""
import collections.abc
import types
from typing import (
    Any,
    Mapping,
    TypeVar,
)

import marsh


_T = TypeVar('_T', bound=Mapping)


@marsh.schema.register
class MappingUnmarshalSchema(marsh.schema.template.MappingUnmarshalSchema[_T]):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        k, v = marsh.utils.inspect_mapping_type(self.value)
        self.key_schema = marsh.schema.UnmarshalSchema(k)
        self.value_schema = marsh.schema.UnmarshalSchema(v)

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~collections.abc.Mapping`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals a mapping type from a mapping input. '
            'Should work for all mapping types. If the mapping '
            'type is one of the abstract mapping classes in '
            '`collections.abc` then a suitable replacement type '
            'is implicitly used instead.'
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return (
                marsh.utils.is_mapping_type(value)
                and issubclass(
                    marsh.utils.get_type(value),
                    collections.abc.Mapping,
                )
            )
        except Exception:
            return False

    def construct(
        self,
        value: Mapping[Any, Any],
    ) -> _T:
        value_type = marsh.utils.get_type(self.value)
        if value_type is collections.abc.MutableMapping:
            return dict(value)  # type: ignore
        if value_type is collections.abc.Mapping:
            return types.MappingProxyType(value)  # type: ignore
        return super().construct(value)



"""Tools that can be used for testing purposes."""
import contextlib
from typing import (
    Any,
    Iterator,
    Optional,
    Type,
    TypeVar,
)

import pytest

import marsh


_T = TypeVar('_T')


@contextlib.contextmanager
def catch_cause(
    base: Type[Exception] = marsh.errors.MarshError,
    cause: Optional[Type[Exception]] = None,
) -> Iterator[None]:
    """Assert that an error is thrown in this context.

    Similar to :func:`pytest.raises` but allows for inspecting
    and finding a specific cause for the error raised by traversing
    the __cause__ attribute chain.

    Arguments:
        base: The base exception class to catch. If not caught,
            the test fails.
        cause: If not :data:`None`, at least one of the exceptions
            in the cause chain must be an instance of this class.

    Returns:
        Context manager.
    """
    with pytest.raises(base) as exc_info:
        yield
    if cause is not None:
        exception_chain = []
        inner_exc: Optional[BaseException] = exc_info.value
        while inner_exc is not None:
            if isinstance(inner_exc, cause):
                return
            exception_chain.append(inner_exc.__class__.__name__)
            inner_exc = inner_exc.__context__
        raise AssertionError(
            f'expected an exception of type {cause} '
            f'as cause, got {tuple(exception_chain)}',
        )


def marshal_succeeds(
    value: Any,
    element: marsh.element.ElementType,
) -> marsh.element.ElementType:
    """Assert that a value can be marshaled.

    The element is also compared to the expected
    output for the marshal.

    Arguments:
        value: The value to marshal.
        element: The expected marshal value.

    Returns:
        The marshaled value if successful.
    """
    schema = marsh.schema.MarshalSchema(value)
    marshaled = schema.marshal()
    element = marsh.element.standardize(element)
    marshaled = marsh.element.standardize(marshaled)
    assert marshaled == element, f'{marshaled} != {element}, schema={schema}'
    return marshaled


def marshal_fails(
    value: Any,
    exception: Optional[Type[Exception]] = None,
) -> None:
    """Assert that a value can not be marshaled.

    Requires that marshaling of the input value
    raises :class:`marsh.errors.MarshalError`.

    Arguments:
        value: Value to marshal.
        exception: Optional cause that must
            exist in the exception chain if
            given.
    """
    with catch_cause(
        base=marsh.errors.MarshalError,
        cause=exception,
    ):
        schema = marsh.schema.MarshalSchema(value)
        schema.marshal()


def unmarshal_succeeds(
    type_: Any,
    element: marsh.element.ElementType,
    value: _T,
) -> _T:
    """Assert that a value is unmarshaled correctly.

    Arguments:
        type_: Unmarshal into this type.
        element: The input that should be unmarshaled.
        value: The expected result.

    Returns:
        The unmarshaled result if successful.
    """
    schema = marsh.schema.UnmarshalSchema[Any](type_)
    unmarshaled = schema.unmarshal(element)
    assert unmarshaled == value, f'{unmarshaled} != {value}'
    return unmarshaled


def unmarshal_fails(
    type_: Any,
    element: marsh.element.ElementType,
    exception: Optional[Type[Exception]] = None,
) -> None:
    """Assert that a value can not be unmarshaled.

    Requires that unmarshaling raises
    :class:`marsh.errors.UnmarshalError`.

    Arguments:
        type_: Attempt to unmarshal into this type.
        element: The input that should be unmarshaled.
        exception: Optional cause that must
            exist in the exception chain if
            given.
    """
    with catch_cause(
        base=marsh.errors.UnmarshalError,
        cause=exception,
    ):
        schema = marsh.schema.UnmarshalSchema[Any](type_)
        schema.unmarshal(element)



from typing import (
    Sequence,
    TypeVar,
)

import marsh
from .. import UnmarshalSchema


_T = TypeVar('_T')


class SequenceUnmarshalSchema(UnmarshalSchema[_T]):
    """A schema for a dynamic sequence of values."""

    value_schema: UnmarshalSchema
    """The schema for the value type."""

    def __str__(
        self,
    ) -> str:
        return f'[{self.value_schema}, ...]'

    def __repr__(
        self,
    ) -> str:
        return (
            f'{super().__repr__()[:-1]}, '
            f'value_schema={repr(self.value_schema)})'
        )

    def doc_field_type(
        self,
    ) -> str:
        return f'[{self.value_schema.doc_field_type()}, ...]'

    def select(
        self,
        path: str,
    ) -> UnmarshalSchema:
        return self.value_schema.select(path)

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _T:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
            return self.construct(())
        if not marsh.utils.is_sequence(element):
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: '
                f'expected sequence element, got: {element}',
            )
        prepared_input = []
        for i, item in enumerate(element):
            with marsh.errors.prepend(i):
                casted_item = self.value_schema.unmarshal(item)
            prepared_input.append(casted_item)
        try:
            return self.construct(prepared_input)
        except marsh.errors.MarshError:
            raise
        except Exception:
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: '
                f'failed to construct sequence',
            )

    def construct(
        self,
        value: Sequence,
    ) -> _T:
        """Build an instance of the type from unmarshaled values.

        Arguments:
            value: The unmarshaled values.

        Returns:
            An instance of the type.
        """
        return marsh.utils.get_type(self.value)(value)



import collections.abc
import numbers
from typing import (
    Any,
    Iterable,
    Iterator,
)

import marsh


BASE_TYPES = (
    float,
    int,
    bool,
    bytes,
    collections.abc.MutableMapping,
    collections.abc.Mapping,
    collections.abc.MutableSequence,
    collections.abc.Sequence,
    str,
    numbers.Complex,
)


def _match_iter(
    protocol: Any,
) -> Iterator:
    try:
        for type_ in BASE_TYPES:
            if issubclass(type_, protocol):
                yield type_
    except TypeError:
        pass


@marsh.schema.caches.new_callable_cache(
    name='marsh.schema.types.protocol.match',
)
def match(
    protocol: Any,
) -> Iterable:
    return marsh.utils.IterableFromIterator(_match_iter(protocol))


@marsh.schema.register
class ProtocolUnmarshalSchema(marsh.schema.UnmarshalSchema[Any]):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.matched = tuple(match(self.value))

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return (
                marsh.utils.is_protocol(value)
                and bool(match(value))
            )
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~typing.Protocol`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Attempts to unmarshal :func:`~typing.runtime_checkable` '
            'Protocols into a matching class. The available classes. '
            'that can be matched against are :class:`float`, '
            ':class:`int`, :class:`bool`, :class:`bytes`, :class:`dict`, '
            ':class:`list`, :class:`str` and :class:`complex`.'
        )

    def doc_field_type(
        self,
    ) -> str:
        return self.value.__name__

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> Any:
        for type_ in self.matched:
            try:
                return marsh.unmarshal(type_, element)
            except Exception:
                pass
        raise marsh.errors.UnmarshalError(
            f'{marsh.utils.get_type_name(self.value)}: failed to '
            f'cast element {element} into one of the classes {self.matched}',
        )



from typing import (
    Any,
    Literal,
    Sequence,
    TypeVar,
    get_args,
)

import marsh


_T = TypeVar('_T')


@marsh.schema.register
class LiteralUnmarshalSchema(marsh.schema.template.UnionUnmarshalSchema[_T]):

    schemas: Sequence['LiteralValueUnmarshalSchema']

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.schemas = tuple(
            LiteralValueUnmarshalSchema(Literal[arg])  # pyright: ignore
            for arg in get_args(self.value)
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return (
            marsh.utils.is_literal(value)
            and len(get_args(value)) > 1
        )

    @staticmethod
    def doc_static_type() -> str:
        return ':data:`~typing.Literal`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Matches the input against the literal value(s), '
            'then returns the matched value.'
        )

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _T:
        try:
            return super().unmarshal(element)
        except Exception:
            closest = marsh.utils.get_closest(
                value=str(element),
                candidates=(schema.literal_value for schema in self.schemas),
            )
            err_msg = (
                'failed to unmarshal input `{element}` into one '
                f'of the literals {self.doc_field_type()}'
            )
            if closest:
                err_msg += f', did you mean "{closest}"?'
            raise marsh.errors.UnmarshalError(err_msg)


@marsh.schema.register
class LiteralValueMarshalSchema(marsh.schema.MarshalSchema):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return (
            marsh.utils.is_literal(value)
            and len(get_args(value)) == 1
        )

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        return get_args(self.value)[0]


@marsh.schema.register
class LiteralValueUnmarshalSchema(marsh.schema.UnmarshalSchema[_T]):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.literal_value = get_args(self.value)[0]
        self.schema: marsh.schema.UnmarshalSchema = \
            marsh.schema.UnmarshalSchema(type(self.literal_value))

    def __str__(
        self,
    ) -> str:
        if isinstance(self.literal_value, str):
            return f'"{self.literal_value}"'
        return str(self.literal_value)

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return (
            marsh.utils.is_literal(value)
            and len(get_args(value)) == 1
        )

    def doc_field_type(
        self,
    ) -> str:
        if isinstance(self.literal_value, str):
            return f'"{self.literal_value}"'
        return str(self.literal_value)

    def doc_description(
        self,
    ) -> str:
        return (
            f'A literal {self.doc_field_type()} of '
            f'of type {self.schema.doc_type()}'
        )

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _T:
        if marsh.utils.is_missing(element):
            if self.has_default():
                return self.get_default()
        literal_value = self.schema.unmarshal(element)
        if literal_value != self.literal_value:
            raise marsh.errors.UnmarshalError(
                f'{marsh.utils.get_type_name(self.value)}: '
                'could not match the literal '
                f'{self.literal_value}: {element}',
            )
        return literal_value



from typing import (
    Any,
    get_args,
)

import marsh
from . import sequence


@marsh.schema.register(lower_priority=sequence.SequenceUnmarshalSchema)
class TupleUnmarshalSchema(marsh.schema.template.StructuredUnmarshalSchema[tuple]):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.schemas = {
            str(i): marsh.schema.UnmarshalSchema(type_)
            for i, type_ in enumerate(
                arg for arg in get_args(self.value)
                if arg != ()  # empty tuple type
            )
        }

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_fixed_size_tuple_type(value)

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`tuple`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals a sequence input into a fixed size '
            'tuple. The input sequence must contain the same '
            'number of elements as the tuple type specifies.'
        )

    def doc_field_type(
        self,
    ) -> str:
        content = ', '.join(
            schema.doc_field_type()
            for schema
            in self.schemas.values()
        )
        return f'[{content}]'

    def construct(
        self,
        *args,
        **kwargs,
    ) -> tuple:
        return tuple(args, **kwargs)



from typing import (
    Any,
    Sequence,
    TypeVar,
)

import marsh


_T = TypeVar('_T')


_PRIORITY = 10


@marsh.schema.register(priority=_PRIORITY)
class NamespaceMarshalSchema(marsh.schema.core.marshal.WrapperMarshalSchema):

    namespaces = marsh.schema.namespace.Namespaces()

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.name = self.namespaces.find_class(marsh.utils.get_type(self.value))

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return bool(cls.namespaces.find_class(marsh.utils.get_type(value)))
        except Exception:
            return False

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        element = self.schema.marshal()
        if not marsh.utils.is_mapping(element):
            raise marsh.errors.MarshalError(
                f'{marsh.utils.get_type_name(self.value)}: marshaling '
                'is not supported for registered component that '
                f'is not marshaled into a mapping element, got {element}',
            )
        if 'name' in element:
            raise marsh.errors.MarshalError(
                'reserved component field `name` encountered',
            )
        return {
            **element,
            **dict(name=self.name),
        }


@marsh.schema.register(priority=_PRIORITY)
class NamespaceUnmarshalSchema(marsh.schema.core.unmarshal.WrapperUnmarshalSchema[_T]):

    namespaces = marsh.schema.namespace.Namespaces()

    def __str__(
        self,
    ) -> str:
        names = self.names()
        type_name = marsh.utils.get_type_name(self.value)
        if names:
            quoted_names = (f'"{name}"' for name in names)
            return f'{type_name}(name={" | ".join(quoted_names)})'
        return type_name

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return bool(cls.namespaces.find_namespaces(value))

    def names(
        self,
    ) -> Sequence[str]:
        return sorted(
            {
                name
                for name, _
                in self.namespaces.find_subclasses(self.value)
            },
        )

    def doc_special_fields(
        self,
    ) -> Sequence[marsh.schema.UnmarshalSchema.Doc.SpecialField]:
        fields = []
        for name, namespace in self.namespaces.find_subclasses(self.value):
            schema = namespace[name]
            field = f'name = "{name}" -> {schema.doc_field_type()}'
            fields.append(
                marsh.schema.UnmarshalSchema.Doc.SpecialField(
                    value=field,
                    description=schema.doc_description(),
                ),
            )
        return fields

    def select(
        self,
        path: str,
    ) -> marsh.schema.UnmarshalSchema:
        if not path:
            return self
        field, path = marsh.path.head(path)
        for name, namespace in self.namespaces.find_subclasses(self.value):
            if name == field:
                return namespace[name].select(path)
        raise marsh.errors.PathError(
            marsh.utils.get_closest_error_message(
                field,
                self.names(),
                key='name',
            ),
        )

    def find_namespace(
        self,
        name: str,
    ) -> marsh.schema.namespace.Namespace:
        """Find namespace for a specified name.

        Raises an error if no namespace is found."""
        candidates = []
        for candidate, namespace in self.namespaces.find_subclasses(self.value):
            if candidate == name:
                return namespace
            candidates.append(candidate)
        raise marsh.errors.UnmarshalError(
            f'{marsh.utils.get_type_name(self.value)}: '
            + marsh.utils.get_closest_error_message(
                value=name,
                candidates=candidates,
                key='name',
            ),
        )

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> _T:
        name: str = marsh.MISSING
        if marsh.utils.is_mapping(element):
            element = dict(element)
            name = element.pop('name', marsh.MISSING)
        if marsh.utils.is_missing(name):
            return self.schema.unmarshal(element)
        return self.find_namespace(name)[name].unmarshal(element)



from typing import Any

import marsh


@marsh.schema.register
class NoneMarshalSchema(marsh.schema.MarshalSchema):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return value in (None, type(None))
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':data:`None`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Simply marshals any :data:`None` value by returning it.'
        )

    def marshal(
        self,
    ) -> None:
        return None


@marsh.schema.register
class NoneUnmarshalSchema(marsh.schema.UnmarshalSchema[None]):

    def __str__(
        self,
    ) -> str:
        return 'None'

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            return value in (None, type(None))
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return ':data:`None`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'A missing value is valid input '
            'as well as the :data:`None` value directly.'
        )

    def doc_type(
        self,
    ) -> str:
        return 'None'

    def doc_field_type(
        self,
    ) -> str:
        return '<None>'

    def doc_description(
        self,
    ) -> str:
        return self.doc_static_description()

    def unmarshal(
        self,
        element: marsh.element.ElementType,
    ) -> None:
        if element is None or marsh.utils.is_missing(element):
            return None
        raise marsh.errors.UnmarshalError(
            f'{marsh.utils.get_type_name(self.value)}: '
            f'expected `None`, got: {element}',
        )



import inspect
from typing import (
    Any,
    Iterable,
    Mapping,
    Sequence,
    TypeVar,
)


import marsh


_T = TypeVar('_T')


GENERIC_PRIORITY = -10


@marsh.schema.register(priority=GENERIC_PRIORITY)
class GenericMappingMarshalSchema(marsh.schema.MarshalSchema):

    value: marsh.utils.MappingProtocol

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~marsh.utils.MappingProtocol`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals mapping values matched through ducktyping.'
            'Uses :class:`~marsh.utils.MappingProtocol` as reference '
            'for the ducktyped match. Marshaling outputs a JSON-style '
            'map.'
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_mapping(value)

    def marshal(
        self,
    ) -> dict:
        return {
            str(key): marsh.marshal(self.value[key])
            for key in self.value.keys()
        }


@marsh.schema.register(priority=GENERIC_PRIORITY)
class GenericSequenceMarshalSchema(marsh.schema.MarshalSchema):

    value: marsh.utils.SequenceProtocol

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~marsh.utils.SequenceProtocol`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals sequence values matched through ducktyping.'
            'Uses :class:`~marsh.utils.SequenceProtocol` as reference '
            'for the ducktyped match. Marshaling outputs a JSON-style '
            'list.'
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_sequence(value)

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        return tuple(
            marsh.marshal(self.value[i])
            for i in range(len(self.value))
        )


@marsh.schema.register(
    priority=GENERIC_PRIORITY,
    higher_priority=(
        GenericMappingMarshalSchema,
        GenericSequenceMarshalSchema,
    ),
)
class GenericIterableMarshalSchema(marsh.schema.MarshalSchema):

    value: Iterable

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        if not marsh.utils.is_obj_instance(value):
            return False
        try:
            iter(value)
            return True
        except TypeError:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return '``Iterable``'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Fetches and marshals all items in an iterable.'
        )

    def marshal(
        self,
    ) -> marsh.element.ElementType:
        return tuple(marsh.marshal(item) for item in self.value)


@marsh.schema.register(priority=GENERIC_PRIORITY)
class GenericCallableUnmarshalSchema(marsh.schema.template.CallableUnmarshalSchema[_T]):

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        try:
            if not marsh.utils.is_callable(value):
                return False
            # make sure value can have signature inspected
            inspect.signature(value)
            return True
        except Exception:
            return False

    @staticmethod
    def doc_static_type() -> str:
        return '``Callable``'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Covers anything that is callable, be it a class, '
            'function or callable class instance. '
            'Callable arguments are inspected and the input '
            'should be a sequence or mapping(similar to '
            '`*args` or `**kwargs`).'
        )


@marsh.schema.register(
    priority=GENERIC_PRIORITY,
    higher_priority=GenericCallableUnmarshalSchema,
)
class GenericMappingUnmarshalSchema(
    marsh.schema.template.MappingUnmarshalSchema[marsh.utils.MappingProtocol],
):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        k, v = marsh.utils.inspect_mapping_type(self.value)
        self.key_schema = marsh.schema.UnmarshalSchema(k)
        self.value_schema = marsh.schema.UnmarshalSchema(v)

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~marsh.utils.MappingProtocol`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals mapping types matched through ducktyping.'
            'Uses :class:`~marsh.utils.MappingProtocol` as reference '
            'for the ducktyped match. Expects input to be a JSON-style '
            'map during unmarshaling.'
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_mapping_type(value)

    def construct(
        self,
        value: Mapping[Any, Any],
    ) -> marsh.utils.MappingProtocol:
        value_type = marsh.utils.get_type(self.value)
        is_abstract = False
        try:
            is_abstract = inspect.isabstract(value_type)
        except Exception:
            pass
        if not is_abstract:
            return super().construct(value)
        return dict(value)


@marsh.schema.register(
    priority=GENERIC_PRIORITY,
    higher_priority=GenericCallableUnmarshalSchema,
)
class GenericSequenceUnmarshalSchema(
    marsh.schema.template.SequenceUnmarshalSchema[marsh.utils.SequenceProtocol],
):

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.value_schema = marsh.schema.UnmarshalSchema(
            marsh.utils.inspect_sequence_type(self.value),
        )

    @staticmethod
    def doc_static_type() -> str:
        return ':class:`~marsh.utils.SequenceProtocol`'

    @staticmethod
    def doc_static_description() -> str:
        return (
            'Unmarshals sequence types matched through ducktyping.'
            'Uses :class:`~marsh.utils.SequenceProtocol` as reference '
            'for the ducktyped match. Expects input to be a JSON-style '
            'list during unmarshaling.'
        )

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        return marsh.utils.is_sequence_type(value)

    def construct(
        self,
        value: Sequence,
    ) -> marsh.utils.SequenceProtocol:
        value_type = marsh.utils.get_type(self.value)
        try:
            is_abstract = inspect.isabstract(value_type)
        except Exception:
            pass
        if is_abstract:
            return list(value)
        return super().construct(value)



"""Collection of general utilities used throughout the framework."""
import ast
import base64
import collections.abc
import dataclasses
import difflib
import functools
import inspect
import os
import shutil
import sys
import weakref
import types
import typing
import typing_extensions
from typing import (
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    Final,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    MutableMapping,
    NamedTuple,
    Optional,
    Protocol,
    Sequence,
    Set,
    SupportsIndex,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    overload,
    runtime_checkable,
)
from typing_extensions import (
    Annotated,
    TypeAlias,
    TypeGuard,
)

import omegaconf
import docstring_parser


_ProtocolTypes: list = []
_AnnotatedTypes: list = []
_LiteralTypes: list = []
_LiteralStringTypes: list = []
_TypeAliasTypes: list = []
_IsTypedDict: list = []


for module in (typing, typing_extensions):
    for attr, container in (
        ('Protocol', _ProtocolTypes),
        ('Annotated', _AnnotatedTypes),
        ('Literal', _LiteralTypes),
        ('LiteralString', _LiteralStringTypes),
        ('TypeAlias', _TypeAliasTypes),
        ('is_typeddict', _IsTypedDict),
    ):
        try:
            container.append(getattr(module, attr))
        except Exception:
            pass


# used by python 3.8 to determine when a typing alias
# is `Annotated`
_AnnotatedInstanceTypes: list = [
    type(ann[None, None])
    for ann
    in _AnnotatedTypes
]


_K = TypeVar('_K')
_K_contra = TypeVar('_K_contra', contravariant=True)
_V = TypeVar('_V')
_V_co = TypeVar('_V_co', covariant=True)
_T = TypeVar('_T')
_C = TypeVar('_C', bound=Callable)
_P = TypeVar('_P', int, float, bool, str)


@runtime_checkable
class SequenceProtocol(Protocol):
    """Protocol matching a sequence class."""

    def __contains__(
        self,
        value,
    ):
        ...

    def __getitem__(
        self,
        index,
    ):
        ...

    def __len__(
        self,
    ) -> int:
        ...


@runtime_checkable
class MappingProtocol(Protocol[_K_contra, _V_co]):
    """Protocol matching a mapping class."""

    def __contains__(
        self,
        key,
    ):
        ...

    def __getitem__(
        self,
        key,
    ):
        ...

    def __iter__(
        self,
    ):
        ...

    @overload
    def get(
        self,
        key: _K_contra,
    ) -> Optional[_V_co]:
        ...

    @overload
    def get(
        self,
        key: _K_contra,
        default: _T,
    ) -> Union[_V_co, _T]:
        ...

    def get(
        self,
        key,
        default=None,
    ):
        ...

    def items(
        self,
    ):
        ...

    def keys(
        self,
    ):
        ...

    def values(
        self,
    ):
        ...


@runtime_checkable
class NamedTupleProtocol(Protocol):
    """Protocol matching a namedtuple class."""

    __doc__: str
    _fields: Sequence[str]
    _field_defaults: Mapping[str, Any]

    def __contains__(
        self,
        value,
    ):
        ...

    def __getitem__(
        self,
        index,
    ):
        ...

    def __len__(
        self,
    ) -> int:
        ...


class SingletonMeta(type):
    """Converts a class into a singleton.

    When the constructor of the class is called
    the first time, an instance is constructed.
    After this, the constructor will always return
    that instance unless it is garbage collected.

    To allow for garbage collection of unused instances
    a weak reference is kept of the instance.
    """

    _instances: MutableMapping[Type[Any], Any] = \
        weakref.WeakKeyDictionary()

    def __call__(
        cls,
        *args,
        **kwargs,
    ) -> None:
        meta = type(cls)
        if cls not in meta._instances:
            meta._instances[cls] = super().__call__(*args, **kwargs)
        return meta._instances[cls]


class _CacheInfoInputType(Protocol):

    @property
    def hits(
        self,
    ) -> int:
        ...

    @property
    def misses(
        self,
    ) -> int:
        ...

    @property
    def maxsize(
        self,
    ) -> Optional[int]:
        ...

    @property
    def currsize(
        self,
    ) -> int:
        ...


class CacheInfo(NamedTuple):
    """Tuple with information for a cache."""
    hits: int = 0
    misses: int = 0
    maxsize: Optional[int] = None
    currsize: int = 0

    @classmethod
    def from_info(
        cls,
        info: _CacheInfoInputType,
    ) -> 'CacheInfo':
        """Create an instance of this class from any cache info
        that follows the same attribute pattern.

        Arguments:
            info: The cache info to use as source

        Returns:
            An instance of this class containing the same
            information as the input.
        """
        return cls(
            hits=info.hits,
            misses=info.misses,
            maxsize=info.maxsize,
            currsize=info.currsize,
        )


class CacheType(Protocol):
    """Represents the protocol for a cache in :mod:`marsh`."""

    def cache_info(
        self,
    ) -> CacheInfo:
        """Get info from the cache.

        Returns:
            The cache info.
        """
        ...

    def cache_clear(
        self,
    ) -> None:
        """Clear the content of the cache."""
        ...

    def cache_disable(
        self,
    ) -> None:
        """Disable the cache."""
        ...

    def cache_enable(
        self,
    ) -> None:
        """Enable the cache."""
        ...


@overload
def cache(
    *,
    maxsize: Optional[int] = None,
    typed: bool = False,
    safe: bool = True,
    binding: Optional[Literal['weak', 'ignore']] = None,
) -> Callable[[_C], _C]:
    ...


@overload
def cache(
    fn: _C,
    /,
    *,
    maxsize: Optional[int] = None,
    typed: bool = False,
    safe: bool = True,
    binding: Optional[Literal['weak', 'ignore']] = None,
) -> _C:
    ...


def cache(
    fn: Optional[_C] = None,
    /,
    *,
    maxsize: Optional[int] = None,
    typed: bool = False,
    safe: bool = True,
    binding: Optional[Literal['weak', 'ignore']] = None,
) -> Union[_C, Callable[[_C], _C]]:
    """Extended LRU cache decorator for functions.

    Allows for safe usage of lru cache, where the :class:`TypeError`
    produced by non-hashable types in the input arguments are caught
    before a new attempt is made to call the function, this time
    bypassing the cache compmletely.

    .. note::

        The bypass is triggered for :class:`TypeError`.
        If the function itself can raise this error it
        will be called twice before the final error is
        propagated.

    The wrapped function fulfills the :class:`CacheType`
    protocol.

    Arguments:
        fn: If :data:`None`, the wrapper is returned,
            otherwise this argument is wrapped and returned.
        maxsize: Maximum size of lru cache.
        typed: Differentiate between different types
            of values that produce the same hash.
        safe: Make a second attempt at calling the function
            and bypassing the cache if the first attempt fails
            with a :class:`TypeError`.
        binding: Should only be used when a method to a class
            is decorated with this function. ``'weak'`` will
            make sure that a weak reference to
            ``self``/``cls``/``metacls`` is stored in the cache
            instead of a strong reference. This allows for objects
            with lru_cache to be garbage collected when no longer
            used. ``'ignore'`` allows the first argument to bypass
            the cache, only performing a lookup on the other arguments.

    Returns:
        The wrapper, unless ``fn`` is given, which instead
        wraps the argument and returns it.
    """
    def wrapper(func: Callable[..., _T]):

        cached_func: Callable[..., _T]
        delegating_func: Callable[..., _T]
        disabled = False

        def cache_disable() -> None:
            nonlocal disabled
            disabled = True

        def cache_enable() -> None:
            nonlocal disabled
            disabled = False

        if binding == 'ignore':
            _self: Any = None

            def cached_func(
                *args,
                **kwargs,
            ) -> _T:
                return func(_self, *args, **kwargs)

            def delegating_func(
                self,
                *args,
                **kwargs,
            ) -> _T:
                if disabled:
                    return func(self, *args, **kwargs)
                nonlocal _self
                _self = self
                return cached_func(*args, **kwargs)

        elif binding == 'weak':

            def cached_func(
                self: weakref.ReferenceType,
                *args,
                **kwargs,
            ) -> _T:
                return func(self(), *args, **kwargs)

            def delegating_func(
                self,
                *args,
                **kwargs,
            ) -> _T:
                if disabled:
                    return func(self, *args, **kwargs)
                return cached_func(weakref.ref(self), *args, **kwargs)

        elif binding is None:
            cached_func = func

            def delegating_func(
                *args,
                **kwargs,
            ) -> _T:
                if disabled:
                    return func(*args, **kwargs)
                return cached_func(*args, **kwargs)

        else:
            raise ValueError(f'unexpected `binding` value: {binding}')
        cached_func = functools.lru_cache(maxsize=maxsize, typed=typed)(cached_func)
        if safe:
            unsafe_delegating_func = delegating_func

            def delegating_func(*args, **kwargs):
                try:
                    return unsafe_delegating_func(*args, **kwargs)
                except TypeError:
                    return func(*args, **kwargs)

        wrapped_func = functools.wraps(func)(delegating_func)
        wrapped_func.__wrapped__ = func  # type: ignore
        wrapped_func.cache_clear = cached_func.cache_clear  # type: ignore
        wrapped_func.cache_info = cached_func.cache_info  # type: ignore
        wrapped_func.cache_disable = cache_disable  # type: ignore
        wrapped_func.cache_enable = cache_enable  # type: ignore
        return wrapped_func
    if fn is None:
        return wrapper
    return wrapper(fn)


class _CachedHashKey(list):
    """Caches the hash value of an item.

    An instance of this class will always
    return this value from ``__hash__``
    without a new hash lookup.

    Arguments:
        value: Hashable value.
    """

    __slots__ = '_hash'

    def __init__(
        self,
        value,
    ):
        self._hash = hash(value)

    def __hash__(self):
        return self._hash


_HASH_KEY_KEYWORD_MARK: Final = (object(),)
_HASH_KEY_FAST_HASH_TYPES: Final = {int, str}


def make_hash_key(
    *args,
    **kwargs,
) -> Any:
    """Make a single cached hashable key from arguments.

    All arguments must be hashable.

    Returns:
        Hashable object that caches its hash, only caclulating it once.
    """
    key = args
    if kwargs:
        key += _HASH_KEY_KEYWORD_MARK
        for item in kwargs.items():
            key += item
    if len(key) == 1 and key[0] in _HASH_KEY_FAST_HASH_TYPES:
        return key[0]
    return _CachedHashKey(key)


def make_typed_hash_key(
    *args,
    **kwargs,
) -> Any:
    """Make a single cached hashable key from arguments.

    Includes the type of each value.

    All arguments must be hashable.

    Returns:
        Hashable object that caches its hash, only caclulating it once.
    """
    key = args + tuple(type(v) for v in args)
    if kwargs:
        key += _HASH_KEY_KEYWORD_MARK
        key += tuple(kwargs.keys())
        key += tuple(kwargs.values())
        key += tuple(type(v) for v in kwargs.values())
    return _CachedHashKey(key)


class SafeDict(MutableMapping[_K, _V]):
    """Dictionary that works with unhashable keys.

    Just like built-in :class:`dict` the order is kept."""

    def __init__(
        self,
    ) -> None:
        self._hashable_cache: Dict[_K, _V] = {}
        self._unhashable_cache: List[Tuple[_K, _V]] = []
        self._order: List[Tuple[bool, int]] = []

    def __getitem__(
        self,
        key: _K,
    ) -> _V:
        try:
            return self._hashable_cache[key]
        except TypeError:
            for k, v in reversed(self._unhashable_cache):
                if k == key:
                    return v
        raise KeyError(key)

    def __setitem__(
        self,
        key: _K,
        value: _V,
    ) -> None:
        try:
            prev_len = len(self._hashable_cache)
            self._hashable_cache[key] = value
            # keep track of order only if there are unhashable
            # keys in this container, otherwise the order is
            # already kept by the underlying dict.
            if not self._unhashable_cache:
                return
            if len(self._hashable_cache) > prev_len:
                self._order.append((True, prev_len))
        except TypeError:
            if not self._unhashable_cache:
                self._order = [(True, i) for i in range(len(self._hashable_cache))]
            for i, (k, _) in enumerate(reversed(self._unhashable_cache), start=1):
                if k == key:
                    self._unhashable_cache[-i] = (key, value)
            else:
                self._order.append((False, len(self._unhashable_cache)))
                self._unhashable_cache.append((key, value))

    def __delitem__(
        self,
        key: _K,
    ) -> None:
        try:
            if key in self._hashable_cache:
                # keep track of order only if there are unhashable
                # keys in this container, otherwise the order is
                # already kept by the underlying dict.
                if self._unhashable_cache:
                    for key_idx, k in enumerate(self._hashable_cache):
                        if k == key:
                            for order_idx, (hashable, idx) in enumerate(self._order):
                                if hashable and idx == key_idx:
                                    self._order.pop(order_idx)
                                    break
                            break
                del self._hashable_cache[key]
                return
        except TypeError:
            for key_idx, (k, _) in enumerate(reversed(self._unhashable_cache), start=1):
                if k == key:
                    self._unhashable_cache.pop(-key_idx)
                    for order_idx, (hashable, idx) in enumerate(self._order):
                        if not hashable and idx == key_idx:
                            self._order.pop(order_idx)
                            break
                    return
        raise KeyError(key)

    def _ordered_iter(
        self,
        reverse: bool,
    ) -> Iterator[_K]:
        if not self._unhashable_cache:
            # use order kept by underlying dict
            if reverse:
                yield from reversed(self._hashable_cache)
            else:
                yield from self._hashable_cache
            return
        hashable_keys = tuple(self._hashable_cache)
        unhashable_keys = tuple(key for key, _ in self._unhashable_cache)
        for hashable, index in (reversed(self._order) if reverse else self._order):
            if hashable:
                yield hashable_keys[index]
            else:
                yield unhashable_keys[index]

    def __iter__(
        self,
    ) -> Iterator[_K]:
        yield from self._ordered_iter(reverse=False)

    def __len__(
        self,
    ) -> int:
        return len(self._hashable_cache) + len(self._unhashable_cache)

    def __reversed__(
        self,
    ) -> Iterator[_K]:
        yield from self._ordered_iter(reverse=True)


class ValueCache(Generic[_K, _V]):
    """A dict-like cache that allows for non-hashable keys.

    The hash for cachable keys are also hashed to improve
    lookup speed.

    Arguments:
        typed: The key type is also taken into consideration
            when matching keys.
    """

    def __init__(
        self,
        typed: bool = False,
    ) -> None:
        self._typed = typed
        self._hits: int = 0
        self._misses: int = 0
        self._disabled = False
        self._cache: SafeDict[Any, _V] = SafeDict()

    def __len__(
        self,
    ) -> int:
        """Get the number of items in the cache.

        Returns:
            Cache size.
        """
        return len(self._cache)

    def __getitem__(
        self,
        key: _K,
    ) -> _V:
        """Get an item from the cache.

        Raises :class:`KeyError` for unmatched keys.

        Arguments:
            key: The key associated with the item to fetch.

        Returns:
            The item.
        """
        if self._disabled:
            raise KeyError(key)
        try:
            self._hits += 1
            return self._cache[self._make_key(key)]
        except KeyError:
            self._hits -= 1
            self._misses += 1
            raise

    def __setitem__(
        self,
        key: _K,
        value: _V,
    ) -> None:
        """Store an item in the cache.

        Arguments:
            key: The key to associate with the item.
            value: The item to store.
        """
        if not self._disabled:
            self._cache[self._make_key(key)] = value

    def __delitem__(
        self,
        key: _K,
    ) -> None:
        """Remove an item from the cache.

        Raises :class:`KeyError` if the given key
        does not exist.

        Arguments:
            key: The key associated with the item.
        """
        del self._cache[self._make_key(key)]

    def _make_key(
        self,
        key: _K,
    ) -> Any:
        try:
            if self._typed:
                return make_typed_hash_key(key)
            return make_hash_key(key)
        except TypeError:
            if self._typed:
                return (key, type(key))
            return key

    def cache_clear(
        self,
    ) -> None:
        """Clear the cache, removing all stored items."""
        self._cache.clear()
        self._hits = self._misses = 0

    def cache_info(
        self,
    ) -> CacheInfo:
        """Get statistics for the cache."""
        return CacheInfo(
            hits=self._hits,
            misses=self._misses,
            currsize=len(self),
        )

    def cache_disable(
        self,
    ) -> None:
        """Disable the cache.

        Getting a value for a key raise :class:`KeyError`.
        Setting a value for a key is a no-op (the value is
        not stored).
        Deleting a value in the cache is still permitted.
        """
        self._disabled = True

    def cache_enable(
        self,
    ) -> None:
        """Enable the cache."""
        self._disabled = False


class CachePool(Mapping[str, CacheType]):
    """A collection of caches."""

    def __init__(
        self,
    ) -> None:
        self._caches: Dict[str, CacheType] = {}

    def __getitem__(
        self,
        key: str,
    ) -> CacheType:
        return self._caches[key]

    def __iter__(
        self,
    ) -> Iterator[str]:
        yield from self._caches

    def __len__(
        self,
    ) -> int:
        return len(self._caches)

    def add(
        self,
        name: str,
        cache: CacheType,
    ) -> None:
        """Add a cache to the collection.

        Arguments:
            name: A name for the cache being added.
            cached: The cache to add.
        """
        self._caches[name] = cache

    @overload
    def new_callable_cache(
        self,
        *,
        name: str,
        maxsize: Optional[int] = None,
        typed: bool = False,
        safe: bool = True,
        binding: Optional[Literal['weak', 'ignore']] = None,
    ) -> Callable[[_C], _C]:
        ...

    @overload
    def new_callable_cache(
        self,
        fn: _C,
        *,
        name: str,
        maxsize: Optional[int] = None,
        typed: bool = False,
        safe: bool = True,
        binding: Optional[Literal['weak', 'ignore']] = None,
    ) -> _C:
        ...

    def new_callable_cache(
        self,
        fn: Optional[_C] = None,
        *,
        name: str,
        maxsize: Optional[int] = None,
        typed: bool = False,
        safe: bool = True,
        binding: Optional[Literal['weak', 'ignore']] = None,
    ) -> Union[_C, Callable[[_C], _C]]:
        """Wrap a function with :func:`cache` while
        simultaneously adding it to this pool.

        See :func:`cache` for descriptions of arguments
        forwarded to the wrapper.

        Arguments:
            name: The name associated with the new cache.

        Returns:
            The cache wrapper.
        """
        def wrapper(func: _C) -> _C:
            wrapped = cache(
                func,
                maxsize=maxsize,
                typed=typed,
                safe=safe,
                binding=binding,
            )
            self.add(
                name=name,
                cache=wrapped,  # type: ignore
            )
            return wrapped
        if fn is None:
            return wrapper
        return wrapper(fn)

    def new_value_cache(
        self,
        name: str,
    ) -> ValueCache:
        """Create a new value cache and add it to
        this pool before returning it.

        Arguments:
            name: The name associated with the new cache.

        Returns:
            The value cache.
        """
        cache: ValueCache = ValueCache()
        self.add(
            name=name,
            cache=cache,
        )
        return cache

    def cache_clear(
        self,
    ) -> None:
        """Clear every cache in this container.

        This does not remove the caches from the container,
        it only calls the ``clear_cache()`` function on
        each cache.
        """
        for cache in self._caches.values():
            cache.cache_clear()

    def cache_info(
        self,
    ) -> Mapping[str, CacheInfo]:
        """Get the cache info for all caches
        in this container.

        Returns:
            A mapping of the name of each cache as key
            and its respective cache info as value.
        """
        return {
            name: cache.cache_info()
            for name, cache
            in self._caches.items()
        }

    def cache_disable(
        self,
    ) -> None:
        """Disable all caches in this container."""
        for cache in self._caches.values():
            cache.cache_disable()

    def cache_enable(
        self,
    ) -> None:
        """Enable all caches in this container."""
        for cache in self._caches.values():
            cache.cache_enable()


class IterableFromIterator(Generic[_T]):
    """Lazily stores results of an iterator.

    Allows for multiple iterations over the
    wrapped iterator while only consuming it
    once.

    Arguments:
        iterator: The iterator to cache.
    """

    def __init__(
        self,
        iterator: Iterator[_T],
    ) -> None:
        self._iter = iterator
        self._cache: list = []

    def __iter__(
        self,
    ) -> Iterator[_T]:
        yield from self._cache
        for item in self._iter:
            self._cache.append(item)
            yield item

    def __bool__(
        self,
    ) -> bool:
        for _ in self:
            return True
        return False


class PriorityOrder(Sequence[_T], Generic[_T]):
    """A mutable sequence of values ordered by priority.

    Higher priority items are moved forward while
    lower priority items are moved backward in the sequence.
    """

    RelativePriority: TypeAlias = Union[None, _V, int, Iterable[_V], Iterable[int]]
    """Relative priority input type."""

    class _RelativePriorityContainer:

        class _PriorityRelation(NamedTuple):
            lower: bool
            higher: bool

        name: str
        ident: int
        lower: Set[int]
        higher: Set[int]

        def __init__(
            self,
            name: str,
            ident: int,
            lower: Set[int],
            higher: Set[int],
        ) -> None:
            if lower.intersection(higher):
                raise ValueError(
                    f'circular priority in {name}',
                )
            self.name = name
            self.ident = ident
            self.lower = lower
            self.higher = higher

        def __lt__(
            self,
            other,
        ) -> bool:
            return self.get_relation(other).lower

        def __gt__(
            self,
            other,
        ) -> bool:
            return self.get_relation(other).higher

        def has_relation(
            self,
            other,
        ) -> bool:
            if not isinstance(other, PriorityOrder._RelativePriorityContainer):
                raise TypeError
            return (
                other.ident in self.lower
                or other.ident in self.higher
                or self.ident in other.lower
                or self.ident in other.higher
            )

        def get_relation(
            self,
            other,
        ) -> 'PriorityOrder._RelativePriorityContainer._PriorityRelation':
            if not isinstance(other, PriorityOrder._RelativePriorityContainer):
                raise TypeError
            lower = (
                self.ident in other.lower
                or other.ident in self.higher
            )
            higher = (
                self.ident in other.higher
                or other.ident in self.lower
            )
            if lower and higher:
                raise ValueError(
                    'circular relative priorities for '
                    f'{self.name} and {other.name}',
                )
            return self._PriorityRelation(
                lower=lower,
                higher=higher,
            )

    class _WrappedValue(Generic[_V]):
        """Wrapes a value, storing it with any contextual information."""

        value: _V
        """The value to wrap"""

        name: str
        """The name of the value"""

        ident: int
        """The hash of the value."""

        priority: int
        """The base priority."""

        relative_priority: 'PriorityOrder._RelativePriorityContainer'
        """The explicitly stated relative priorities."""

        index: int
        """Used to keep track of in what order values were added."""

        _counter: ClassVar[int] = 0
        """Keeps track of the current count which is used to set
        the `index` value."""

        def __init__(
            self,
            value: _V,
            priority: int,
            lower_priority: 'PriorityOrder.RelativePriority[_V]',
            higher_priority: 'PriorityOrder.RelativePriority[_V]',
        ) -> None:
            self.value = value
            self.name = str(value)
            self.ident = hash(value)
            self.priority = priority
            self.relative_priority = PriorityOrder._RelativePriorityContainer(
                name=self.name,
                ident=self.ident,
                lower=self.resolve_priorities(lower_priority),
                higher=self.resolve_priorities(higher_priority),
            )
            self.index = self.__class__._counter
            self.__class__._counter += 1

        def __str__(
            self,
        ) -> str:
            return self.name

        @staticmethod
        def resolve_priorities(
            priority: 'PriorityOrder.RelativePriority[_V]',
        ) -> Set[int]:
            if priority is None:
                return set()
            try:
                if (
                    isinstance(priority, collections.abc.Iterable)
                    and not isinstance(priority, (str, bytes))
                ):
                    return set(map(hash, priority))
            except Exception:
                pass
            return {hash(priority)}

    def __init__(
        self,
    ) -> None:
        self._values: List[PriorityOrder._WrappedValue[_T]] = []

    def __len__(
        self,
    ) -> int:
        return len(self._values)

    @overload
    def __getitem__(
        self,
        index: SupportsIndex,
    ) -> _T:
        ...

    @overload
    def __getitem__(
        self,
        index: slice,
    ) -> 'PriorityOrder[_T]':
        ...

    def __getitem__(
        self,
        index: Union[SupportsIndex, slice],
    ) -> Union[_T, 'PriorityOrder[_T]']:
        """Get a slice from this container or one of its values."""
        if isinstance(index, slice):
            new = self.__class__()
            new._values = self._values[index]
            return new
        return self._values[index].value

    def __delitem__(
        self,
        index: Union[SupportsIndex, slice],
    ) -> None:
        """Delete one or more values in this container.

        If the deleted value(s) had any relative priority relationship
        with the remaining values the entire priority order is recalculated."""
        if isinstance(index, slice):
            removed_values = self._values[index]
        else:
            removed_values = [self._values[index]]
        del self._values[index]
        # If removed value(s) had any relative priority relation
        # with current values recalculate the entire order instead
        # of doing something more fancy
        for value in self._values:
            for removed_value in removed_values:
                if value.relative_priority.has_relation(removed_value.relative_priority):
                    self.reload()
                    return

    def reload(
        self,
    ) -> None:
        """Reshuffle values in this container based on priority."""
        values = sorted(self._values, key=lambda x: x.index)
        del self._values[:]
        for value in values:
            self.add(
                value=value.value,
                priority=value.priority,
                lower_priority=value.relative_priority.lower,
                higher_priority=value.relative_priority.higher,
            )

    @overload
    def add(
        self,
        value: _T,
        *,
        priority: int = 0,
        lower_priority: 'PriorityOrder.RelativePriority[_T]' = None,
        higher_priority: 'PriorityOrder.RelativePriority[_T]' = None,
        replace: bool = False,
    ) -> _T:
        ...

    @overload
    def add(
        self,
        *,
        priority: int = 0,
        lower_priority: 'PriorityOrder.RelativePriority[_T]' = None,
        higher_priority: 'PriorityOrder.RelativePriority[_T]' = None,
        replace: bool = False,
    ) -> Callable[[_T], _T]:
        ...

    def add(
        self,
        value: Optional[_T] = None,
        *,
        priority: int = 0,
        lower_priority: 'PriorityOrder.RelativePriority[_T]' = None,
        higher_priority: 'PriorityOrder.RelativePriority[_T]' = None,
        replace: bool = False,
    ) -> Union[Callable[[_T], _T], _T]:

        def wrapper(value: _T) -> _T:
            wrapped = self._WrappedValue(
                value=value,
                priority=priority,
                lower_priority=lower_priority,
                higher_priority=higher_priority,
            )

            # does the value to be added already exist?
            for i in range(len(self._values)):
                if self._values[i] == wrapped:
                    if replace:
                        del self._values[i]
                        break
                    raise ValueError(
                        f'{wrapped} already exists.',
                    )

            # find insertion index for value based on priorities.
            # circular priorities with a length larger than 1 do
            # not raise an error.
            idx = 0
            for i, position in enumerate(self._values):
                if wrapped.relative_priority > position.relative_priority:
                    break
                if (
                    wrapped.relative_priority < position.relative_priority
                    or wrapped.priority < position.priority
                ):
                    idx = i + 1
            self._values.insert(idx, wrapped)
            return value
        if value is None:
            return wrapper
        return wrapper(value)


class WeakTypeCache:
    """A type cache that holds its values as weak references.

    When iterated over the types are yielded in reverse order,
    the most recently added is yielded first.

    Arguments:
        skip_types: Any types to ignore. If attempting
            to store one of these types it becomes a no-op.
    """

    def __init__(
        self,
        skip_types: Iterable[Any] = (),
    ) -> None:
        self._types: MutableMapping[weakref.ReferenceType, None] = \
            weakref.WeakKeyDictionary()
        self._skip_types = set(skip_types)

    def __len__(
        self,
    ) -> int:
        return len(self._types)

    def __iter__(
        self,
    ) -> Iterator[Any]:
        data = getattr(self._types, 'data', None)
        if isinstance(data, dict):
            for ref in reversed(data):
                type_ = ref()
                if type_ is not None:
                    yield type_
        else:
            yield from reversed(tuple(self._types))

    def __contains__(
        self,
        type_: Any,
    ) -> bool:
        return type_ in self._types

    def add(
        self,
        type_: Any,
    ) -> None:
        """Add a type to the cache.

        Arguments:
            type_: The type to add.
        """
        if is_typing_alias(type_):
            return
        if type_ in self._skip_types:
            return
        try:
            del self._types[type_]
        except KeyError:
            pass
        self._types[type_] = None

    def clear(
        self,
    ) -> None:
        """Clear the cache, removing all stored types."""
        self._types.clear()


def is_missing(
    value: Any,
) -> bool:
    """Determine if a value is missing.

    Supports :attr:`marsh.MISSING`, :attr:`omegaconf.MISSING`
    and :attr:`dataclasses.MISSING`.

    Arguments:
        value: Any value to test for missing.

    Returns:
        ``True`` if missing, else ``False``.
    """
    return value in (omegaconf.MISSING, dataclasses.MISSING)


def is_sequence(
    value: Any,
) -> TypeGuard[Sequence]:
    """Determine if a value is a sequence (excluding string, bytes).

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a sequence, else ``False``.
    """
    try:
        return (
            not is_mapping(value)
            and not isinstance(value, (str, bytes))
            and isinstance(value, SequenceProtocol)
            and is_obj_instance(value)
        )
    except TypeError:
        return False


def is_sequence_type(
    value: Any,
) -> TypeGuard[Type[Sequence]]:
    """Determine if a value is a sequence type (excluding string, bytes types).

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a sequence type, else ``False``.
    """
    if (origin := get_origin(value)) is not None:
        return is_sequence_type(origin)
    try:
        return (
            not is_mapping_type(value)
            and not issubclass(value, (str, bytes))
            and isinstance(value, SequenceProtocol)
            and not is_obj_instance(value)
        )
    except TypeError:
        return False


def is_empty_tuple_type(
    value: Any,
) -> TypeGuard[Type[Tuple[()]]]:
    """Determine if a value is an empty tuple type
    (``typing.Tuple[()]``, ``tuple[()]``).

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a an empty tuple type, else ``False``.
    """
    # only type aliases can specify contents of tuple
    if get_origin(value) is not tuple:
        return False
    # type alias without arguments is considered Tuple[Any, ...]
    if value is Tuple:
        return False
    args = get_args(value)
    # Empty args is considered empty tuple in 3.9+ when using
    # tuple[()] instead of typing.Tuple[()]
    if not args:
        return True
    # For typing.Tuple[()] the empty tuple argument is retained.
    return args == ((),)


def is_fixed_size_tuple_type(
    value,
) -> bool:
    """Determine if a value is fixed size tuple type.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a a fixed size tuple type, else ``False``.
    """
    return (
        (
            value is tuple
            or get_origin(value) is tuple
        )
        and (
            (
                (args := get_args(value))
                and ... not in args
            )
            or is_empty_tuple_type(value)
        )
    )


def is_mapping(
    value: Any,
) -> TypeGuard[Mapping]:
    """Determine if a value is a mapping.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a mapping, else ``False``.
    """
    try:
        return (
            isinstance(value, MappingProtocol)
            and is_obj_instance(value)
        )
    except TypeError:
        return False


def is_mapping_type(
    value: Any,
) -> TypeGuard[Type[Mapping]]:
    """Determine if a value is a mapping type.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a mapping, else ``False``.
    """
    if (origin := get_origin(value)) is not None:
        return is_mapping_type(origin)
    try:
        return (
            isinstance(value, MappingProtocol)
            and not is_obj_instance(value)
        )
    except TypeError:
        return False


def is_primitive(
    value: Any,
) -> TypeGuard[Union[int, float, bool, str]]:
    """Determine if a value is primitive.

    ``PrimitiveElement: int | float | bool | str``

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is primitive, else ``False``.
    """
    try:
        return isinstance(value, (int, float, str, bool))
    except TypeError:
        return False


def is_primitive_type(
    value: Any,
) -> TypeGuard[Union[Type[int], Type[float], Type[bool], Type[str]]]:
    """Determine if a value is primitive type.

    ``PrimitiveElement: int | float | bool | str``

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a primitive type, else ``False``.
    """
    try:
        return issubclass(value, (int, float, str, bool))
    except TypeError:
        return False


def is_namedtuple(
    value: Any,
) -> TypeGuard[NamedTupleProtocol]:
    """Determine if a value is a namedtuple.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a namedtuple, else ``False``.
    """
    return (
        is_obj_instance(value)
        and is_namedtuple_type(get_type(value))
    )


def is_namedtuple_type(
    value: Any,
) -> TypeGuard[Type[NamedTupleProtocol]]:
    """Determine if a value is a namedtuple type.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a namedtuple type, else ``False``.
    """
    try:
        if not isinstance(value, NamedTupleProtocol):
            return False
        for cls in inspect.getmro(value):  # type: ignore
            if tuple in cls.__bases__:
                return True
    except Exception:
        pass
    return False


def is_typed_namedtuple(
    value: Any,
) -> TypeGuard[NamedTupleProtocol]:
    """Determine if a value is a typed namedtuple.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a namedtuple, else ``False``.
    """
    try:
        return (
            is_namedtuple(value)
            and bool(get_type_hints(value))
        )
    except TypeError:
        return False


def is_typed_namedtuple_type(
    value: Any,
) -> TypeGuard[Type[NamedTupleProtocol]]:
    """Determine if a value is a typed namedtuple type.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a typed namedtuple type, else ``False``.
    """
    try:
        return (
            is_namedtuple_type(value)
            and bool(get_type_hints(value))
        )
    except TypeError:
        return False


def is_typeddict_type(
    value: Any,
) -> bool:
    """Determine if a value is a typed dict type.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a typed dict type, else ``False``.
    """
    return any(is_td(value) for is_td in _IsTypedDict)


def is_defaultdict_type(
    value: Any,
) -> bool:
    """Determine if a value is a default dict type.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a default dict type, else ``False``.
    """
    if (
        value is DefaultDict
        or get_origin(value) is collections.defaultdict
    ):
        return True
    try:
        return issubclass(value, collections.defaultdict)
    except TypeError:
        return False


def is_callable(
    value: Any,
) -> bool:
    """Determine if a value is callable.

    Always returns ``False`` if the value is
    a protocol or typing alias.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a callable,
        else ``False``.
    """
    try:
        return (
            callable(value)
            and not is_protocol(value)
            and not is_typing_alias(value)
        )
    except Exception:
        return False


def is_obj_instance(
    value: Any,
) -> bool:
    """Determine if a value is an instantiated object.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is an iinstantiated object,
        else ``False``.
    """
    if is_typing_alias(value):
        return False
    try:
        return not issubclass(type(value), type)
    except Exception:
        return False


def is_annotated(
    value: Any,
) -> bool:
    """Determine if a value originates from :data:`typing.Annotated`.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value originates from :data:`typing.Annotated`, else ``False``.
    """
    return (
        get_origin(value) is Annotated  # python 3.9+
        or typing_extensions.get_origin(value) is Annotated  # python 3.8
    )


def get_annotations(
    value: Any,
) -> Sequence:
    """Extract the additional arguments of an instance of :data:`typing.Annotated`

    Arguments:
        value: The :data:`typing.Annotated` instance.

    Returns:
        The annotations.
    """
    if not is_annotated(value):
        raise ValueError(
            f'expected `Annotated` type, got {value}',
        )
    if get_origin(value) is Annotated:
        return get_args(value)[1:]
    return typing_extensions.get_args(value)[1:]


def is_protocol(
    value: Any,
) -> bool:
    """Determine if a value is a protocol.

    Any class that has :class:`typing.Protocol` in its
    immediate bases is considered a protocol.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a protocol, else ``False``.
    """
    for cls in getattr(value, '__bases__', ()):
        if cls in _ProtocolTypes:
            return True
    return False


def is_literal(
    value: Any,
) -> bool:
    """Determine if a value is a typing.Literal.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a typing.Literal, else ``False``.
    """
    return get_origin(value) in _LiteralTypes


def is_literal_string(
    value: Any,
) -> bool:
    """Determine if a value is a typing.LiteralString.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a typing.LiteralString, else ``False``.
    """
    return value in _LiteralStringTypes


def is_typing_alias(
    value: Any,
) -> bool:
    """Determine if a value is a typing alias.

    If the class name ends with ``'_GenericAlias'``,
    ``'_SpecialForm'`` or ``'_AnnotatedAlias'``
    the value is considered a typing alias.

    For python 3.9+ a check is made against :class:`types.GenericAlias`.

    Arguments:
        value: Any value to evaluate.

    Returns:
        ``True`` if value is a typing alias, else ``False``.
    """
    return (
        get_origin(value) is not None
        or type(value) is getattr(types, 'GenericAlias', None)
        or value in _TypeAliasTypes
        or hasattr(value, '__class__')
        and (
            value.__class__.__name__.endswith('_GenericAlias')
            or value.__class__.__name__.endswith('_SpecialForm')
            or value.__class__.__name__.endswith('_AnnotatedAlias')
        )
    )


def is_optional(
    type_: Any,
) -> bool:
    """Determine if a type is a optional.

    Arguments:
        type: The type to evaluate.

    Returns:
        ``True`` if value is ``typing.Union`` and
        :data:`None` is in its arguments, else ``False``.
    """
    if is_annotated(type_):
        type_ = get_origin(type_)
    return get_origin(type_) is Union \
        and type(None) in get_args(type_)


def get_type(
    value: Any,
) -> Any:
    """Attempt to retrieve the type of an object.

    If the value is a generic typing alias, the origin type is returned.
    I.e. ``get_type(typing.List[str])`` returns the builtins ``list`` type.

    If the value is an instance of a class, the class is returned.

    If the value is a class, it is directly returned.

    >>> assert get_type([1, 2, 3]) == get_type(list)

    Arguments:
        value: Value to retrieve type from.

    Returns:
        The type.
    """
    if (origin := get_origin(value)) is not None:
        return get_type(origin)
    if is_typing_alias(value):
        return value
    try:
        if issubclass(type(value), type):
            return value
    except Exception:
        pass
    return type(value)


def get_optional_type(
    type_: Any,
) -> Any:
    """Get the non-optional version of
    an optional type.

    Arguments:
        type: The optional type.

    Returns:
        The type as non-optional.
    """
    if not is_optional(type_):
        raise ValueError(
            'attempted to fetch optional type '
            'from non-optional type variable: '
            f'{type_}',
        )
    if is_annotated(type_):
        type_ = get_origin(type_)
    args = []
    for arg in get_args(type_):
        if arg not in (None, type(None)):  # noqa: E721
            args.append(arg)
    if len(args) == 1:
        return args[0]
    return Union[tuple(args)]  # pyright: ignore


def get_type_name(
    value: Any,
) -> str:
    """Get a string representation for the type of a value.

    Arguments:
        value: A type or instance of a type to get
            a string representation for.

    Returns:
        The string representation of the value.
    """
    if is_annotated(value):
        value = get_origin(value)
    if is_typing_alias(value):
        return str(value).replace('typing.', '')
    return getattr(get_type(value), '__name__', str(value))


def bytes_to_base64(
    value: bytes,
) -> str:
    """Encode bytes to a base64 string.

    Follows the format used by protobuf.

    Arguments:
        value: The bytes to encode.

    Returns:
        The base64 string representing the bytes argument.
    """
    return base64.urlsafe_b64encode(value).decode('utf-8')


def base64_to_bytes(
    value: str,
) -> bytes:
    """Decode a base64 string to bytes.

    Follows the format used by protobuf.

    Arguments:
        value: The base64 string to decode.

    Returns:
        The bytes representation of the base64 string argument.
    """
    encoded = value.encode('utf-8')
    padded_value = encoded + b'=' * (4 - len(encoded) % 4)
    return base64.urlsafe_b64decode(padded_value)


def primitive_to_bool(
    value: Union[int, float, bool, str],
) -> bool:
    """Convert a primitive value to a bool.

    Follows the YAML convention instead of
    using ``__bool__`` function of the object,
    see :func:`str_to_bool` for more info.

    Raises :class:`ValueError` on failure.

    Arguments:
        value: The primitive value to convert
            into bool.

    Returns:
        The bool representation of the value argument.
    """
    if isinstance(value, bool):
        return value
    if value == 1:
        return True
    elif value == 0:
        return False
    return str_to_bool(str(value))


def str_to_bool(
    value: str,
) -> bool:
    """Cast a string to a boolean.

    ``True: 1 | [Tt][Rr][Uu][Ee] | [Yy] | [Yy][Ee][Ss] | [Oo][Nn]``

    ``False: 0 | [Ff][Aa][Ll][Ss][Ee] | [Nn] | [Nn][Oo] | [Oo][Ff][Ff]``

    Inspired by https://yaml.org/type/bool.html

    Raises :class:`ValueError` if the string does not
    conform to this format.

    Arguments:
        value: String to cast.

    Returns:
        The input string parsed as a bool.
    """
    if value.lower() in ('1', 'true', 'y', 'yes', 'on'):
        return True
    elif value.lower() in ('0', 'false', 'n', 'no', 'off'):
        return False
    raise ValueError(
        'Expected 1 | 0 | [Tt][Rr][Uu][Ee] | '
        '[Yy] | [Yy][Ee][Ss] | [Oo][Nn] | '
        '[Ff][Aa][Ll][Ss][Ee] | [Nn] | '
        f'[Nn][Oo] | [Oo][Oo][Ff], got: {value}',
    )


def str_to_int(
    value: str,
) -> int:
    """Cast a string to an int.

    Compared to using :class:`int` to cast the value
    this function allows values such as ``"0.0"`` by
    first parsing the value as a float before casting
    it to an int (unless it contains decimals).

    Raises :class:`ValueError` if the input string can not be
    parsed as an int.

    Arguments:
        value: String input to parse as int.

    Returns:
        The parsed int.
    """
    try:
        return int(value)
    except ValueError as e:
        try:
            return float_to_int(float(value))
        except ValueError:
            raise e from None


def float_to_int(
    value: float,
) -> int:
    """Cast a float to an int.

    Raises :class:`ValueError` if the input float is not
    an integer (contains decimal values).
    This differs from using :class:`int`
    to cast a float since decimal values
    produce an error instead of being dropped.

    Raises :class:`ValueError` on failure.

    Arguments:
        value: Float input to cast as int.

    Returns:
        Integer of the input value.
    """
    if not value.is_integer():
        raise ValueError(
            f'could not convert to int: {value}',
        )
    return int(value)


def cast_primitive(
    type_: Type[_P],
    value: Union[int, float, bool, str],
) -> _P:
    """Cast a primitive value to a specified primitive type.

    Uses :func:`primitive_to_bool`, :func:`str_to_int`
    and :func:`float_to_int` when applicable.

    Raises :class:`ValueError` if unsuccessful.

    Arguments:
        primitive: The primitive type (:class:`int`, :class:`float`,
            :class:`bool` or :class:`str`).
        value: A primitive value to cast.

    Returns:
        The casted value.
    """
    if not issubclass(type_, (int, float, bool, str)):
        raise ValueError(f'only primitive (sub)types are valid: {type_}')
    if not is_primitive(value):
        raise ValueError(f'only primitive values are valid: {value}')
    if issubclass(type_, bool):
        return primitive_to_bool(value)
    if issubclass(type_, int) and isinstance(value, str):
        return str_to_int(value)
    if issubclass(type_, int) and isinstance(value, float):
        return float_to_int(value)
    return type_(value)


def cast_literal(
    literal: _P,
    value: Union[int, float, bool, str],
) -> _P:
    """Attempts to cast a primtive value to the same type as a primitive
    literal and match it.

    Raises :class:`ValueError` if not able to match.

    Arguments:
        literal: The primitive literal to cast to.
        value: The value to cast.

    Returns:
        The literal.
    """
    if cast_primitive(type(literal), value) != literal:
        raise ValueError(f'could not convert to {literal}: {value}')
    return literal


def match_literal(
    literals: Iterable[Union[int, float, bool, str]],
    value: Union[int, float, bool, str],
) -> Union[int, float, bool, str]:
    """Attempts to cast and match a value to a set of primitive literals.

    Raises :class:`ValueError` if not able to match.

    Arguments:
        literals: The primitive literals to match.
        value: The value to cast and match.

    Returns:
        The literal that was matched.
    """
    for literal in literals:
        try:
            return cast_literal(
                literal=literal,  # type: ignore
                value=value,
            )
        except ValueError:
            pass
    raise ValueError(
        'could not match any of the literals '
        f'{tuple(literals)}: {value}',
    )


def cast_none(
    value: Any,
) -> None:
    """Attempt to cast a value to :data:`None`.

    Allowed values are case insensitive
    strings ``"null"`` and ``"none"``
    or a missing value.

    Raises :class:`ValueError` on failure.

    Arguments:
        value: The value to cast to :data:`None`

    Returns:
        The casted value.
    """
    if (
        is_missing(value)
        or value in (None, type(None))
    ):
        return None
    if str(value).lower() in ('null', 'none'):
        return None
    raise ValueError(f'could not cast to `None`: {value}')


def extract_description(
    doc: str,
) -> str:
    """Omit all but the main text in a docstring.

    Attempts to omit things such as arguments and
    return value from a docstring.

    Arguments:
        doc: Docstring.

    Returns:
        The description if found, else the entire docstring.
    """
    description = ''
    if not doc:
        return description
    try:
        parsed = docstring_parser.parse(doc)
        try:
            if parsed.short_description:
                description = parsed.short_description
        except Exception:
            pass
        if parsed.long_description:
            if description:
                description = f'{description}\n\n{parsed.long_description}'
            else:
                description = parsed.long_description
    except Exception:
        pass
    return description or doc


def get_description(
    cls,
) -> Optional[str]:
    """Get the docstring for a class.

    An attempt is made to omit all but the
    main text in the docstring. This means
    that the return value should be missing
    things such as arguments and return value
    descriptions.

    Arguments:
        cls: Class to retrieve description from.

    Returns:
        The description if found, else :data:`None`.
    """
    doc = getattr(cls, '__doc__', None)
    if not doc:
        return None
    return extract_description(doc)


def _get_attribute_description(
    cls,
    name: str,
) -> Optional[str]:
    doc = getattr(cls, '__doc__', None)
    if doc:
        try:
            for param in docstring_parser.parse(doc).params:
                try:
                    if param.arg_name == name and param.description:
                        return param.description
                except Exception:
                    pass
        except Exception:
            pass
    try:
        ast_body = ast.parse(inspect.getsource(cls)).body[0].body  # type: ignore
    except Exception:
        return None
    node_iter = iter(ast_body)
    while True:
        try:
            node = next(node_iter)
        except StopIteration:
            break
        if (
            isinstance(node, ast.AnnAssign)
            and isinstance(node.target, ast.Name)
            and node.target.id == name
        ):
            try:
                node = next(node_iter)
                if not isinstance(node, ast.Expr):
                    continue
                if isinstance(node.value, ast.Constant):
                    if isinstance(node.value.value, str):
                        return inspect.cleandoc(node.value.value)
            except StopIteration:
                pass
            break
    return None


def get_attribute_description(
    cls,
    name: str,
) -> Optional[str]:
    """Get the docstring for an attribute of
    a class.

    Arguments:
        cls: Class for which an attribute description
            should be fetched.
        name: The name of the attribute.

    Returns:
        The attribute description if found, else :data:`None`.
    """
    try:
        for base in inspect.getmro(cls):
            if base == object:
                continue
            descr = _get_attribute_description(
                base,
                name=name,
            )
            if descr:
                return descr
    except Exception:
        try:
            return _get_attribute_description(
                cls,
                name=name,
            )
        except Exception:
            pass
    return None


def as_dataclass(
    cls: Type[_T],
) -> Type[_T]:
    """Safely convert a class into a dataclass.

    This function can be called multiple times on the same
    class without it raising an error the second time.

    Arguments:
        cls: Class to convert into dataclass.

    Returns:
        Dataclass.
    """
    assert not is_obj_instance(cls)
    if not hasattr(cls, '__dataclass_fields__'):
        return dataclasses.dataclass(cls)
    for field in dataclasses.fields(cls):
        if (
            field.name in getattr(cls, '__annotations__', {})
            and field.default_factory != dataclasses.MISSING
        ):
            setattr(cls, field.name, field)
    return dataclasses.dataclass(cls)


def get_closest(
    value: str,
    candidates: Iterable[str],
) -> Optional[str]:
    """Get the closest value to some input from a set of candidates.

    If no close value is found, :data:`None` is returned.

    Arguments:
        value: The input value.
        candidates: The values to evaluate closeness to.

    Returns:
        The closest candidate. :data:`None` if no candidate is close.
    """
    close = difflib.get_close_matches(
        value,
        candidates,
        n=1,
    )
    if close:
        return close[0]
    return None


def get_closest_error_message(
    value: str,
    candidates: Iterable[str],
    key: str = 'key',
) -> str:
    """Get an error message for a key.

    The error message can take two forms, the
    first being ``'unknown "{key}"'`` and the
    second being ``'unknown "{key}", did you mean "{closest}"?'``
    which is only returned when a close match is found
    among the candidates.

    Arguments:
        value: The input value.
        candidates: The values to evaluate closeness to.
        key: A name for the key value.

    Returns:
        The error message.
    """
    message = f'unknown {key} "{value}"'
    closest = get_closest(
        value=value,
        candidates=candidates,
    )
    if closest:
        message += f', did you mean "{closest}"?'
    return message


def inspect_sequence_type(
    sequence_type: Any,
) -> Any:
    """Get the type of the content of a sequence type.

    Expects a single content type.
    If no content type is found or there are multiple content
    types ``typing.Any`` is returned.

    Arguments:
        type_: The type variable of a sequence.

    Returns:
        The type for the content of the input sequence type.

    >>> get_sequence_type(typing.List[int])
    <class 'int'>
    >>> get_sequence_type(typing.Tuple[str, ...])
    <class 'str'>
    >>> get_sequence_type(list)
    typing.Any
    >>> get_sequence_type(typing.Tuple[str, int, float])
    typing.Any
    """
    if is_annotated(sequence_type):
        sequence_type = get_origin(sequence_type)
    orig_bases = getattr(
        sequence_type,
        '__orig_bases__',
        (),
    )
    for base in (sequence_type,) + orig_bases:
        if is_sequence_type(base):
            t_args = get_args(base)
            if (
                len(t_args) == 1
                or len(t_args) == 2 and ... in t_args
            ):
                for t in t_args:
                    if t != ...:
                        return t
    return Any


class _MappingKeyValueTypesInfo(NamedTuple):
    key: Any
    value: Any


def inspect_mapping_type(
    mapping_type: Any,
) -> _MappingKeyValueTypesInfo:
    """Get the key and value type from a mapping type.

    Returns ``typing.Any`` for types that could not be inferred.

    Arguments:
        type_: The type variable of a mapping.

    Returns:
        The key and value type.

    >>> get_sequence_type(typing.Dict[str, int])
    (<class 'str'>, <class 'int'>)
    >>> get_sequence_type(typing.Mapping[str, typing.List[int]]])
    (<class 'str'>, typing.List[int])
    >>> get_sequence_type(dict)
    (typing.Any, typing.Any)
    """
    if is_annotated(mapping_type):
        mapping_type = get_origin(mapping_type)
    orig_bases = getattr(
        mapping_type,
        '__orig_bases__',
        getattr(
            get_type(mapping_type),
            '__orig_bases__',
            (),
        ),
    )
    for base in (mapping_type, get_type(mapping_type)) + orig_bases:
        if is_mapping_type(base):
            t_args = get_args(base)
            if len(t_args) == 2:
                return _MappingKeyValueTypesInfo(t_args[0], t_args[1])
    return _MappingKeyValueTypesInfo(Any, Any)


def is_testing() -> bool:
    """Discover if the current runtime is for testing with Pytest."""
    return (
        'PYTEST_CURRENT_TEST' in os.environ
        or 'pytest' in sys.modules
    )


def get_terminal_width() -> int:
    """Attempt to get the width of the current terminal.

    The terminal size is preprocessed in a similar way
    to :mod:`argparse`.
    """
    return max(
        shutil.get_terminal_size().columns - 2,
        11,
    )



import re
import os
from setuptools import (
    setup,
    find_packages,
)
from typing import Optional


def find_version(
    fpath: str,
) -> Optional[str]:
    with open(fpath, 'r') as fp:
        match = re.search(
            r'(?<=__version__ = [\'"])([^\'"]+)(?=[\'"])',
            fp.read(),
        )
    if not match:
        return None
    return match.group(1)


root = os.path.dirname(os.path.abspath(__file__))


with open(os.path.join(root, 'README.md'), 'r') as fp:
    long_description = fp.read()


version = find_version(os.path.join(root, 'marsh/__init__.py'))
if not version:
    raise RuntimeError('could not find version of marsh')


setup(
    name='marsh',
    version=version,
    description='(Un)marshaling framework.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/adriansahlman/marsh',
    author='Adrian Sahlman',
    author_email='adrian.sahlman@gmail.com',
    license='MIT',
    packages=find_packages(include=('marsh', 'marsh.*')),
    package_data={'marsh': ['py.typed']},
    install_requires=(
        'typing-extensions >= 4.2',
        'omegaconf >= 2.1.1',
        'docstring-parser >= 0.7.3',
        'dateparser >= 1.1.1',
    ),
    python_requires='>=3.8',
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
    ],
    keywords='marshal unmarshal configuration',
)



"""A set of template implementations of schemas.

These templates are useful as base classes when extending
the framework to support more types."""
from .callable import CallableUnmarshalSchema
from .mapping import MappingUnmarshalSchema
from .sequence import SequenceUnmarshalSchema
from .structured import StructuredUnmarshalSchema
from .union import UnionUnmarshalSchema


__all__ = (
    'CallableUnmarshalSchema',
    'MappingUnmarshalSchema',
    'SequenceUnmarshalSchema',
    'StructuredUnmarshalSchema',
    'UnionUnmarshalSchema',
)



from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    SupportsIndex,
    Type,
    TypeVar,
    Union,
    overload,
    cast,
)

import marsh


_S = TypeVar('_S', bound='Schema')


class SchemaSelection(List[Type[_S]], Generic[_S]):
    """A sequence of schema types that were matched against a value.

    Contains one or more schema types, with all but the last one
    being wrapper classes."""

    @overload
    def __getitem__(
        self,
        index: SupportsIndex,
    ) -> Type[_S]:
        ...

    @overload
    def __getitem__(
        self,
        index: slice,
    ) -> 'SchemaSelection[_S]':
        ...

    def __getitem__(
        self,
        index: Union[SupportsIndex, slice],
    ) -> Union[Type[_S], 'SchemaSelection[_S]']:
        if isinstance(index, slice):
            return self.__class__(super().__getitem__(index))
        return super().__getitem__(index)

    def __add__(
        self,
        other,
    ) -> 'SchemaSelection[_S]':
        return self.__class__(list(self) + list(other))

    def build(
        self,
        *args,
        **kwargs,
    ) -> _S:
        """Build the selected schema(s) into a single schema.

        Arguments:
            args: Any positional arguments to pass
                on to the schema constructor.
            kwargs: Any keyword arguments to pass
                on to the schema constructor.

        Returns:
            Schema instance.
        """
        if len(self) > 1:
            return self[0](
                *args,
                wrapped=self[1:],
                **kwargs,
            )
        return self[0](
            *args,
            **kwargs,
        )


class SchemaRegistry(Generic[_S]):
    """Holds schema classes in a prioritized order.

    Arguments:
        error: Type of the error to throw
            when no schema could be matched
            during selection.
    """

    priority: marsh.utils.PriorityOrder[Type[_S]]

    def __init__(
        self,
        error: Type[Exception] = marsh.errors.MarshError,
    ) -> None:
        self.priority = marsh.utils.PriorityOrder()
        self.error = error

    def __iter__(
        self,
    ) -> Iterator[Type[_S]]:
        yield from self.priority

    @overload
    def register(
        self,
        schema: Type[_S],
        *,
        priority: int = 0,
        lower_priority: marsh.utils.PriorityOrder.RelativePriority[Type[_S]] = None,
        higher_priority: marsh.utils.PriorityOrder.RelativePriority[Type[_S]] = None,
        replace: bool = False,
    ) -> Type[_S]:
        ...

    @overload
    def register(
        self,
        *,
        priority: int = 0,
        lower_priority: marsh.utils.PriorityOrder.RelativePriority[Type[_S]] = None,
        higher_priority: marsh.utils.PriorityOrder.RelativePriority[Type[_S]] = None,
        replace: bool = False,
    ) -> Callable[[Type[_S]], Type[_S]]:
        ...

    def register(
        self,
        schema: Optional[Type[_S]] = None,
        *,
        priority: int = 0,
        lower_priority: marsh.utils.PriorityOrder.RelativePriority[Type[_S]] = None,
        higher_priority: marsh.utils.PriorityOrder.RelativePriority[Type[_S]] = None,
        replace: bool = False,
    ) -> Union[Callable[[Type[_S]], Type[_S]], Type[_S]]:
        if schema is not None:
            return self.priority.add(
                schema,
                priority=priority,
                lower_priority=lower_priority,
                higher_priority=higher_priority,
                replace=replace,
            )
        return self.priority.add(
            priority=priority,
            lower_priority=lower_priority,
            higher_priority=higher_priority,
            replace=replace,
        )

    def match(
        self,
        value: Any,
    ) -> SchemaSelection[_S]:
        """Match a value to one or more schemas types.

        Selected schema types are returned in the order
        they were match. The matching is canceled and
        the results are returned when the first non-wrapper
        schema type has been matched.

        Raises an error if no non-wrapper schema could be matched.

        Arguments:
            value: Value to match against schema types.

        Returns:
            Selection of schema types.
        """
        schemas: SchemaSelection[_S] = SchemaSelection()
        for schema in self.priority:
            try:
                if schema.match(value):
                    schemas.append(schema)
                    if not issubclass(schema, WrapperSchema):
                        return schemas
            except TypeError:
                pass
        raise self.error(
            f'no schema matched value: {value}',
        )


class SchemaMeta(type, Generic[_S]):
    """Metaclass for schemas.

    If calling the constructor on the base schema class
    it will instead match the input with all registered
    schemas and return an instance of a subclass."""

    registry: SchemaRegistry[_S] = SchemaRegistry()
    """The registered schema types."""

    def __call__(
        cls,
        value,
        *args,
        **kwargs,
    ):
        if cls is not Schema:
            return type(object).__call__(cls, value, *args, **kwargs)
        return cls.registry.match(value).build(value, *args, **kwargs)

    def __init_subclass__(
        metacls,
    ) -> None:
        super().__init_subclass__()
        if hasattr(metacls, 'registry') and metacls.registry is not Schema.registry:
            return
        metacls.registry = SchemaRegistry()


class Schema(metaclass=SchemaMeta):
    """Base class for schemas.

    A schema is a class that can be matched
    against a value and then initialized with
    that value and any other arguments.

    The base class works as a factory when
    its called as a constructor, selecting
    the correctly matched schema and returning
    an instance of it.

    Arguments:
        value: The value to match.
        args: Unused but caught positional arguments.
        kwargs: Unused but caught keyword arguments.
    """

    def __init__(
        self,
        value: Any,
        *args,
        **kwargs,
    ) -> None:
        self.value = value

    @classmethod
    def match(
        cls,
        value: Any,
    ) -> bool:
        """Match this class against a value.

        Arguments:
            value: The value to match against.

        Returns:
            ``True`` if matched, else ``False``.
        """
        raise NotImplementedError

    @staticmethod
    def doc_static_type() -> Optional[str]:
        """Get a static string representation
        of the type supported by this schema.

        Returns:
            The name of the type.
        """
        return None

    @staticmethod
    def doc_static_description() -> Optional[str]:
        """Get a static string description
        of how the type supported by this schema
        is used.

        Returns:
            The description.
        """
        return None


class WrapperSchema(Schema, Generic[_S]):
    """Wraps another schema, potentially
    manipulating calls and return values to/from it.

    The wrapper will lazily initialize its wrapped schema,
    holding its type and input arguments instead of an
    instance of the schema up until the first time
    the schema is accessed.

    Arguments:
        args: Positional arguments stored and passed
            to the constructor of the schema being
            wrapped.
        wrapped: A schema type or an iterable of schema
            types in wrapping order.
        kwargs: Keyword arguments stored and passed
            to the constructor of the schema being
            wrapped.
    """

    wrapped: Sequence[Type[_S]]
    """The types of the schemas that are wrapped."""

    @property
    def schema(
        self,
    ) -> _S:
        """Lazy reference to inner schema. Constructed the first time it is
        being accessed."""
        if getattr(self, '_schema', None) is None:
            if not (
                hasattr(self, 'wrapped')
                and hasattr(self, '_args')
                and hasattr(self, '_kwargs')
            ):
                raise AttributeError(
                    'can not access attribute "schema" before '
                    'WrapperSchema.__init__() call',
                )
            wrapped = self.wrapped[1:]
            schema_type = self.wrapped[0]
            if marsh.utils.is_missing(schema_type):
                raise ValueError(
                    f'{self.__class__.__name__}: wrapped internal '
                    'schema type is missing',
                )
            if wrapped:
                self._schema = schema_type(
                    *self._args,
                    wrapped=wrapped,
                    **self._kwargs,
                )
            else:
                self._schema = schema_type(
                    *self._args,
                    **self._kwargs,
                )
        return self._schema

    @schema.setter
    def schema(
        self,
        value: _S,
    ) -> None:
        self._schema = value

    @property
    def inner_schema(
        self,
    ) -> _S:
        """Bypass any wrapper succeeding wrapper schemas
        to extract the most inner schema."""
        schema = self.schema
        while isinstance(schema, WrapperSchema):
            schema = schema.schema
        return schema

    _args: tuple
    """Positional arguments which are passed to the schema type
    being wrapped when constructed."""

    _kwargs: Mapping
    """Keyword arguments which are passed to the schema type
    being wrapped when constructed."""

    def __init__(
        self,
        *args,
        wrapped: Union[Type[_S], Iterable[Type[_S]]],
        **kwargs,
    ) -> None:
        try:
            if issubclass(wrapped, Schema):  # type: ignore
                wrapped = (cast(Type[_S], wrapped),)
            else:
                wrapped = tuple(cast(Iterable[Type[_S]], wrapped))
        except Exception:
            wrapped = tuple(cast(Iterable[Type[_S]], wrapped))
        if not wrapped:
            raise ValueError(
                f'{self.__class__.__name__}: at least one schema '
                'type must be given to a wrapper schema',
            )
        if issubclass(wrapped[-1], WrapperSchema):
            raise ValueError(
                f'{self.__class__.__name__}: the final schema type '
                'being wrapped may not be a wrapper itself: '
                f'{wrapped[-1].__name__}',
            )
        super().__init__(*args, **kwargs)
        self.wrapped = wrapped
        self._args = args
        self._kwargs = kwargs


caches: marsh.utils.CachePool = marsh.utils.CachePool()
"""Pool of caches used throughout :mod:`marsh.schema`"""
